import functools
import weakref

import sentry_sdk
from sentry_sdk.api import continue_trace
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations import DidNotEnable, Integration, _check_minimum_version
from sentry_sdk.integrations.logging import ignore_logger
from sentry_sdk.scope import Scope, should_send_default_pii
from sentry_sdk.traces import SegmentNameSource
from sentry_sdk.tracing import TransactionSource
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import (
    SENSITIVE_DATA_SUBSTITUTE,
    capture_internal_exceptions,
    event_from_exception,
    format_timestamp,
    parse_version,
)

try:
    from rq.job import JobStatus
    from rq.queue import Queue
    from rq.timeouts import JobTimeoutException
    from rq.version import VERSION as RQ_VERSION
    from rq.worker import Worker
except ImportError:
    raise DidNotEnable("RQ not installed")

try:
    from rq.worker import BaseWorker

    if not hasattr(BaseWorker, "perform_job"):
        BaseWorker = None
except ImportError:
    BaseWorker = None

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable

    from rq.job import Job

    from sentry_sdk._types import Event, EventProcessor
    from sentry_sdk.utils import ExcInfo


class RqIntegration(Integration):
    identifier = "rq"
    origin = f"auto.queue.{identifier}"

    @staticmethod
    def setup_once() -> None:
        version = parse_version(RQ_VERSION)
        _check_minimum_version(RqIntegration, version)

        # In rq 2.7.0+, SimpleWorker inherits from BaseWorker directly
        # instead of Worker, so we need to patch BaseWorker to cover both.
        # For older versions where BaseWorker doesn't exist or doesn't have
        # perform_job, we patch Worker.
        worker_cls = BaseWorker if BaseWorker is not None else Worker

        old_perform_job = worker_cls.perform_job

        @functools.wraps(old_perform_job)
        def sentry_patched_perform_job(
            self: "Any", job: "Job", queue: "Queue", *args: "Any", **kwargs: "Any"
        ) -> bool:
            client = sentry_sdk.get_client()
            if client.get_integration(RqIntegration) is None:
                return old_perform_job(self, job, queue, *args, **kwargs)

            with sentry_sdk.new_scope() as scope:
                scope.clear_breadcrumbs()
                scope.add_event_processor(_make_event_processor(weakref.ref(job)))

                if has_span_streaming_enabled(client.options):
                    sentry_sdk.traces.continue_trace(
                        job.meta.get("_sentry_trace_headers") or {}
                    )

                    Scope.set_custom_sampling_context({"rq_job": job})

                    func_name = None
                    with capture_internal_exceptions():
                        func_name = job.func_name

                    with sentry_sdk.traces.start_span(
                        name="unknown RQ task" if func_name is None else func_name,
                        attributes={
                            "sentry.op": OP.QUEUE_TASK_RQ,
                            "sentry.origin": RqIntegration.origin,
                            "sentry.segment.name.source": SegmentNameSource.TASK,
                            SPANDATA.MESSAGING_MESSAGE_ID: job.id,
                            SPANDATA.MESSAGING_DESTINATION_NAME: queue.name,
                        },
                        parent_span=None,
                    ) as span:
                        if func_name is not None:
                            span.set_attribute(SPANDATA.CODE_FUNCTION_NAME, func_name)

                        rv = old_perform_job(self, job, queue, *args, **kwargs)
                else:
                    transaction = continue_trace(
                        job.meta.get("_sentry_trace_headers") or {},
                        op=OP.QUEUE_TASK_RQ,
                        name="unknown RQ task",
                        source=TransactionSource.TASK,
                        origin=RqIntegration.origin,
                    )

                    with capture_internal_exceptions():
                        transaction.name = job.func_name

                    with sentry_sdk.start_transaction(
                        transaction,
                        custom_sampling_context={"rq_job": job},
                    ) as span:
                        span.set_data(SPANDATA.MESSAGING_DESTINATION_NAME, queue.name)

                        rv = old_perform_job(self, job, queue, *args, **kwargs)

            if self.is_horse:
                # We're inside of a forked process and RQ is
                # about to call `os._exit`. Make sure that our
                # events get sent out.
                sentry_sdk.get_client().flush()

            return rv

        worker_cls.perform_job = sentry_patched_perform_job

        old_handle_exception = worker_cls.handle_exception

        def sentry_patched_handle_exception(
            self: "Worker", job: "Any", *exc_info: "Any", **kwargs: "Any"
        ) -> "Any":
            retry = (
                hasattr(job, "retries_left")
                and job.retries_left
                and job.retries_left > 0
            )
            failed = job._status == JobStatus.FAILED or job.is_failed
            if failed and not retry:
                _capture_exception(exc_info)

            return old_handle_exception(self, job, *exc_info, **kwargs)

        worker_cls.handle_exception = sentry_patched_handle_exception

        old_enqueue_job = Queue.enqueue_job

        @functools.wraps(old_enqueue_job)
        def sentry_patched_enqueue_job(
            self: "Queue", job: "Any", **kwargs: "Any"
        ) -> "Any":
            client = sentry_sdk.get_client()
            if client.get_integration(RqIntegration) is None:
                return old_enqueue_job(self, job, **kwargs)

            scope = sentry_sdk.get_current_scope()
            span = (
                scope.streamed_span
                if has_span_streaming_enabled(client.options)
                else scope.span
            )
            if span is not None:
                job.meta["_sentry_trace_headers"] = dict(
                    scope.iter_trace_propagation_headers()
                )

            return old_enqueue_job(self, job, **kwargs)

        Queue.enqueue_job = sentry_patched_enqueue_job

        ignore_logger("rq.worker")


def _make_event_processor(weak_job: "Callable[[], Job]") -> "EventProcessor":
    def event_processor(event: "Event", hint: "dict[str, Any]") -> "Event":
        job = weak_job()
        if job is not None:
            with capture_internal_exceptions():
                extra = event.setdefault("extra", {})
                rq_job = {
                    "job_id": job.id,
                    "func": job.func_name,
                    "args": (
                        job.args
                        if should_send_default_pii()
                        else SENSITIVE_DATA_SUBSTITUTE
                    ),
                    "kwargs": (
                        job.kwargs
                        if should_send_default_pii()
                        else SENSITIVE_DATA_SUBSTITUTE
                    ),
                    "description": job.description,
                }

                if job.enqueued_at:
                    rq_job["enqueued_at"] = format_timestamp(job.enqueued_at)
                if job.started_at:
                    rq_job["started_at"] = format_timestamp(job.started_at)

                extra["rq-job"] = rq_job

        if "exc_info" in hint:
            with capture_internal_exceptions():
                if issubclass(hint["exc_info"][0], JobTimeoutException):
                    event["fingerprint"] = ["rq", "JobTimeoutException", job.func_name]

        return event

    return event_processor


def _capture_exception(exc_info: "ExcInfo", **kwargs: "Any") -> None:
    client = sentry_sdk.get_client()

    event, hint = event_from_exception(
        exc_info,
        client_options=client.options,
        mechanism={"type": "rq", "handled": False},
    )

    sentry_sdk.capture_event(event, hint=hint)



from unittest import mock

import pytest

import sentry_sdk
from sentry_sdk.tracing_utils import Baggage


@pytest.mark.parametrize("sample_rand", (0.0, 0.25, 0.5, 0.75))
@pytest.mark.parametrize("sample_rate", (0.0, 0.25, 0.5, 0.75, 1.0))
def test_deterministic_sampled(sentry_init, capture_events, sample_rate, sample_rand):
    """
    Test that sample_rand is generated on new traces, that it is used to
    make the sampling decision, and that it is included in the transaction's
    baggage.
    """
    sentry_init(traces_sample_rate=sample_rate)
    events = capture_events()

    with mock.patch(
        "sentry_sdk.tracing_utils.Random.randrange",
        return_value=int(sample_rand * 1000000),
    ):
        with sentry_sdk.start_transaction() as transaction:
            assert (
                transaction.get_baggage().sentry_items["sample_rand"]
                == f"{sample_rand:.6f}"  # noqa: E231
            )

    # Transaction event captured if sample_rand < sample_rate, indicating that
    # sample_rand is used to make the sampling decision.
    assert len(events) == int(sample_rand < sample_rate)


@pytest.mark.parametrize("sample_rand", (0.0, 0.25, 0.5, 0.75))
@pytest.mark.parametrize("sample_rate", (0.0, 0.25, 0.5, 0.75, 1.0))
def test_transaction_uses_incoming_sample_rand(
    sentry_init, capture_events, sample_rate, sample_rand
):
    """
    Test that the transaction uses the sample_rand value from the incoming baggage.
    """
    baggage = Baggage(sentry_items={"sample_rand": f"{sample_rand:.6f}"})  # noqa: E231

    sentry_init(traces_sample_rate=sample_rate)
    events = capture_events()

    with sentry_sdk.start_transaction(baggage=baggage) as transaction:
        assert (
            transaction.get_baggage().sentry_items["sample_rand"]
            == f"{sample_rand:.6f}"  # noqa: E231
        )

    # Transaction event captured if sample_rand < sample_rate, indicating that
    # sample_rand is used to make the sampling decision.
    assert len(events) == int(sample_rand < sample_rate)



from unittest import mock

import pytest
from fakeredis import FakeStrictRedis

import sentry_sdk
from sentry_sdk import capture_message, start_transaction
from sentry_sdk.consts import SPANDATA
from sentry_sdk.integrations.redis import RedisIntegration

MOCK_CONNECTION_POOL = mock.MagicMock()
MOCK_CONNECTION_POOL.connection_kwargs = {
    "host": "localhost",
    "port": 63791,
    "db": 1,
}


def test_basic(sentry_init, capture_events):
    sentry_init(integrations=[RedisIntegration()])
    events = capture_events()

    connection = FakeStrictRedis()

    connection.get("foobar")
    capture_message("hi")

    (event,) = events
    (crumb,) = event["breadcrumbs"]["values"]

    assert crumb == {
        "category": "redis",
        "message": "GET 'foobar'",
        "data": {
            "redis.key": "foobar",
            "redis.command": "GET",
            "redis.is_cluster": False,
            "db.operation": "GET",
        },
        "timestamp": crumb["timestamp"],
        "type": "redis",
    }


@pytest.mark.parametrize("span_streaming", [True, False])
@pytest.mark.parametrize(
    "is_transaction, send_default_pii, expected_first_ten",
    [
        (False, False, ["GET 'foo'", "SET 'bar' [Filtered]", "SET 'baz' [Filtered]"]),
        (True, True, ["GET 'foo'", "SET 'bar' 1", "SET 'baz' 2"]),
    ],
)
def test_redis_pipeline(
    sentry_init,
    capture_events,
    capture_items,
    is_transaction,
    send_default_pii,
    expected_first_ten,
    span_streaming,
):
    sentry_init(
        integrations=[RedisIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    connection = FakeStrictRedis()

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            pipeline = connection.pipeline(transaction=is_transaction)
            pipeline.get("foo")
            pipeline.set("bar", 1)
            pipeline.set("baz", 2)
            pipeline.execute()
        sentry_sdk.flush()

        assert len(items) == 2
        pipeline_span, parent_span = items[0].payload, items[1].payload

        assert parent_span["name"] == "custom parent"
        assert parent_span["is_segment"] is True

        assert pipeline_span["name"] == "redis.pipeline.execute"
        assert pipeline_span["attributes"]["sentry.op"] == "db.redis"
        assert pipeline_span["attributes"]["sentry.origin"] == "auto.db.redis"
        assert pipeline_span["attributes"][SPANDATA.DB_SYSTEM_NAME] == "redis"
    else:
        events = capture_events()
        with start_transaction():
            pipeline = connection.pipeline(transaction=is_transaction)
            pipeline.get("foo")
            pipeline.set("bar", 1)
            pipeline.set("baz", 2)
            pipeline.execute()

        (event,) = events
        (span,) = event["spans"]
        assert span["op"] == "db.redis"
        assert span["description"] == "redis.pipeline.execute"
        assert span["data"][SPANDATA.DB_SYSTEM] == "redis"
        assert span["data"]["redis.commands"] == {
            "count": 3,
            "first_ten": expected_first_ten,
        }
        assert span["tags"] == {
            "redis.transaction": is_transaction,
            "redis.is_cluster": False,
        }


@pytest.mark.parametrize("span_streaming", [True, False])
def test_sensitive_data(sentry_init, capture_events, capture_items, span_streaming):
    # fakeredis does not support the AUTH command, so we need to mock it
    with mock.patch(
        "sentry_sdk.integrations.redis.utils._COMMANDS_INCLUDING_SENSITIVE_DATA",
        ["get"],
    ):
        sentry_init(
            integrations=[RedisIntegration()],
            traces_sample_rate=1.0,
            send_default_pii=True,
            trace_lifecycle="stream" if span_streaming else "static",
        )

        connection = FakeStrictRedis()

        if span_streaming:
            items = capture_items("span")
            with sentry_sdk.traces.start_span(name="custom parent"):
                connection.get("this is super secret")
            sentry_sdk.flush()

            assert len(items) == 2
            redis_span, parent_span = items[0].payload, items[1].payload

            assert parent_span["name"] == "custom parent"
            assert redis_span["name"] == "GET [Filtered]"
            assert redis_span["attributes"][SPANDATA.DB_QUERY_TEXT] == "GET [Filtered]"
            assert redis_span["attributes"]["sentry.op"] == "db.redis"
        else:
            events = capture_events()
            with start_transaction():
                connection.get(
                    "this is super secret"
                )  # because fakeredis does not support AUTH we use GET instead

            (event,) = events
            spans = event["spans"]
            assert spans[0]["op"] == "db.redis"
            assert spans[0]["description"] == "GET [Filtered]"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_pii_data_redacted(sentry_init, capture_events, capture_items, span_streaming):
    sentry_init(
        integrations=[RedisIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    connection = FakeStrictRedis()

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            connection.set("somekey1", "my secret string1")
            connection.set("somekey2", "my secret string2")
            connection.get("somekey2")
            connection.delete("somekey1", "somekey2")
        sentry_sdk.flush()

        assert len(items) == 5
        set1, set2, get, delete, parent = [item.payload for item in items]

        assert parent["name"] == "custom parent"
        assert set1["name"] == "SET 'somekey1' [Filtered]"
        assert set1["attributes"][SPANDATA.DB_QUERY_TEXT] == "SET 'somekey1' [Filtered]"
        assert set1["attributes"]["sentry.op"] == "db.redis"
        assert set2["name"] == "SET 'somekey2' [Filtered]"
        assert set2["attributes"][SPANDATA.DB_QUERY_TEXT] == "SET 'somekey2' [Filtered]"
        assert get["name"] == "GET 'somekey2'"
        assert delete["name"] == "DEL 'somekey1' [Filtered]"
    else:
        events = capture_events()
        with start_transaction():
            connection.set("somekey1", "my secret string1")
            connection.set("somekey2", "my secret string2")
            connection.get("somekey2")
            connection.delete("somekey1", "somekey2")

        (event,) = events
        spans = event["spans"]
        assert spans[0]["op"] == "db.redis"
        assert spans[0]["description"] == "SET 'somekey1' [Filtered]"
        assert spans[1]["description"] == "SET 'somekey2' [Filtered]"
        assert spans[2]["description"] == "GET 'somekey2'"
        assert spans[3]["description"] == "DEL 'somekey1' [Filtered]"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_pii_data_sent(sentry_init, capture_events, capture_items, span_streaming):
    sentry_init(
        integrations=[RedisIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=True,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    connection = FakeStrictRedis()

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            connection.set("somekey1", "my secret string1")
            connection.set("somekey2", "my secret string2")
            connection.get("somekey2")
            connection.delete("somekey1", "somekey2")
        sentry_sdk.flush()

        assert len(items) == 5
        set1, set2, get, delete, parent = [item.payload for item in items]

        assert parent["name"] == "custom parent"
        assert set1["name"] == "SET 'somekey1' 'my secret string1'"
        assert (
            set1["attributes"][SPANDATA.DB_QUERY_TEXT]
            == "SET 'somekey1' 'my secret string1'"
        )
        assert set1["attributes"]["sentry.op"] == "db.redis"
        assert set2["name"] == "SET 'somekey2' 'my secret string2'"
        assert (
            set2["attributes"][SPANDATA.DB_QUERY_TEXT]
            == "SET 'somekey2' 'my secret string2'"
        )
        assert get["name"] == "GET 'somekey2'"
        assert delete["name"] == "DEL 'somekey1' 'somekey2'"
    else:
        events = capture_events()
        with start_transaction():
            connection.set("somekey1", "my secret string1")
            connection.set("somekey2", "my secret string2")
            connection.get("somekey2")
            connection.delete("somekey1", "somekey2")

        (event,) = events
        spans = event["spans"]
        assert spans[0]["op"] == "db.redis"
        assert spans[0]["description"] == "SET 'somekey1' 'my secret string1'"
        assert spans[1]["description"] == "SET 'somekey2' 'my secret string2'"
        assert spans[2]["description"] == "GET 'somekey2'"
        assert spans[3]["description"] == "DEL 'somekey1' 'somekey2'"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_no_data_truncation_by_default(
    sentry_init, capture_events, capture_items, span_streaming
):
    sentry_init(
        integrations=[RedisIntegration()],
        traces_sample_rate=1.0,
        send_default_pii=True,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    connection = FakeStrictRedis()
    long_string = "a" * 100000
    short_string = "b" * 10

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            connection.set("somekey1", long_string)
            connection.set("somekey2", short_string)
        sentry_sdk.flush()

        assert len(items) == 3
        set1, set2, parent = [item.payload for item in items]

        assert parent["name"] == "custom parent"
        assert set1["name"] == f"SET 'somekey1' '{long_string}'"
        assert (
            set1["attributes"][SPANDATA.DB_QUERY_TEXT]
            == f"SET 'somekey1' '{long_string}'"
        )
        assert set1["attributes"]["sentry.op"] == "db.redis"
        assert set2["name"] == f"SET 'somekey2' '{short_string}'"
        assert (
            set2["attributes"][SPANDATA.DB_QUERY_TEXT]
            == f"SET 'somekey2' '{short_string}'"
        )
    else:
        events = capture_events()
        with start_transaction():
            connection.set("somekey1", long_string)
            connection.set("somekey2", short_string)

        (event,) = events
        spans = event["spans"]
        assert spans[0]["op"] == "db.redis"
        assert spans[0]["description"] == f"SET 'somekey1' '{long_string}'"
        assert spans[1]["description"] == f"SET 'somekey2' '{short_string}'"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_data_truncation_custom(
    sentry_init, capture_events, capture_items, span_streaming
):
    sentry_init(
        integrations=[RedisIntegration(max_data_size=30)],
        traces_sample_rate=1.0,
        send_default_pii=True,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    connection = FakeStrictRedis()
    long_string = "a" * 100000
    short_string = "b" * 10
    expected_long = "SET 'somekey1' '%s..." % (
        long_string[: 30 - len("...") - len("SET 'somekey1' '")],
    )
    expected_short = "SET 'somekey2' '%s'" % (short_string,)

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            connection.set("somekey1", long_string)
            connection.set("somekey2", short_string)
        sentry_sdk.flush()

        assert len(items) == 3
        set1, set2, parent = [item.payload for item in items]

        assert parent["name"] == "custom parent"
        assert set1["name"] == expected_long
        assert (
            set1["attributes"][SPANDATA.DB_QUERY_TEXT]
            == f"SET 'somekey1' '{long_string}'"
        )
        assert set1["attributes"]["sentry.op"] == "db.redis"
        assert set2["name"] == expected_short
        assert (
            set2["attributes"][SPANDATA.DB_QUERY_TEXT]
            == f"SET 'somekey2' '{short_string}'"
        )
    else:
        events = capture_events()
        with start_transaction():
            connection.set("somekey1", long_string)
            connection.set("somekey2", short_string)

        (event,) = events
        spans = event["spans"]
        assert spans[0]["op"] == "db.redis"
        assert spans[0]["description"] == expected_long
        assert spans[1]["description"] == expected_short


def test_breadcrumbs(sentry_init, capture_events):
    sentry_init(
        integrations=[RedisIntegration(max_data_size=30)],
        send_default_pii=True,
    )
    events = capture_events()

    connection = FakeStrictRedis()

    long_string = "a" * 100000
    connection.set("somekey1", long_string)
    short_string = "b" * 10
    connection.set("somekey2", short_string)

    capture_message("hi")

    (event,) = events
    crumbs = event["breadcrumbs"]["values"]

    assert crumbs[0] == {
        "message": "SET 'somekey1' 'aaaaaaaaaaa...",
        "type": "redis",
        "category": "redis",
        "data": {
            "db.operation": "SET",
            "redis.is_cluster": False,
            "redis.command": "SET",
            "redis.key": "somekey1",
        },
        "timestamp": crumbs[0]["timestamp"],
    }
    assert crumbs[1] == {
        "message": "SET 'somekey2' 'bbbbbbbbbb'",
        "type": "redis",
        "category": "redis",
        "data": {
            "db.operation": "SET",
            "redis.is_cluster": False,
            "redis.command": "SET",
            "redis.key": "somekey2",
        },
        "timestamp": crumbs[1]["timestamp"],
    }


@pytest.mark.parametrize("span_streaming", [True, False])
def test_db_connection_attributes_client(
    sentry_init, capture_events, capture_items, span_streaming
):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[RedisIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            connection = FakeStrictRedis(connection_pool=MOCK_CONNECTION_POOL)
            connection.get("foobar")
        sentry_sdk.flush()

        assert len(items) == 2
        redis_span, parent_span = items[0].payload, items[1].payload

        assert parent_span["name"] == "custom parent"
        assert redis_span["name"] == "GET 'foobar'"
        attrs = redis_span["attributes"]
        assert attrs["sentry.op"] == "db.redis"
        assert attrs[SPANDATA.DB_QUERY_TEXT] == "GET 'foobar'"
        assert attrs[SPANDATA.DB_SYSTEM_NAME] == "redis"
        assert attrs[SPANDATA.DB_DRIVER_NAME] == "redis-py"
        assert attrs[SPANDATA.DB_NAMESPACE] == "1"
        assert attrs[SPANDATA.SERVER_ADDRESS] == "localhost"
        assert attrs[SPANDATA.SERVER_PORT] == 63791
    else:
        events = capture_events()
        with start_transaction():
            connection = FakeStrictRedis(connection_pool=MOCK_CONNECTION_POOL)
            connection.get("foobar")

        (event,) = events
        (span,) = event["spans"]

        assert span["op"] == "db.redis"
        assert span["description"] == "GET 'foobar'"
        assert span["data"][SPANDATA.DB_SYSTEM] == "redis"
        assert span["data"][SPANDATA.DB_DRIVER_NAME] == "redis-py"
        assert span["data"][SPANDATA.DB_NAME] == "1"
        assert span["data"][SPANDATA.SERVER_ADDRESS] == "localhost"
        assert span["data"][SPANDATA.SERVER_PORT] == 63791


@pytest.mark.parametrize("span_streaming", [True, False])
def test_db_connection_attributes_pipeline(
    sentry_init, capture_events, capture_items, span_streaming
):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[RedisIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            connection = FakeStrictRedis(connection_pool=MOCK_CONNECTION_POOL)
            pipeline = connection.pipeline(transaction=False)
            pipeline.get("foo")
            pipeline.set("bar", 1)
            pipeline.set("baz", 2)
            pipeline.execute()
        sentry_sdk.flush()

        assert len(items) == 2
        pipeline_span, parent_span = items[0].payload, items[1].payload

        assert parent_span["name"] == "custom parent"
        assert pipeline_span["name"] == "redis.pipeline.execute"
        attrs = pipeline_span["attributes"]
        assert attrs["sentry.op"] == "db.redis"
        assert attrs[SPANDATA.DB_SYSTEM_NAME] == "redis"
        assert attrs[SPANDATA.DB_DRIVER_NAME] == "redis-py"
        assert attrs[SPANDATA.DB_NAMESPACE] == "1"
        assert attrs[SPANDATA.SERVER_ADDRESS] == "localhost"
        assert attrs[SPANDATA.SERVER_PORT] == 63791
    else:
        events = capture_events()
        with start_transaction():
            connection = FakeStrictRedis(connection_pool=MOCK_CONNECTION_POOL)
            pipeline = connection.pipeline(transaction=False)
            pipeline.get("foo")
            pipeline.set("bar", 1)
            pipeline.set("baz", 2)
            pipeline.execute()

        (event,) = events
        (span,) = event["spans"]

        assert span["op"] == "db.redis"
        assert span["description"] == "redis.pipeline.execute"
        assert span["data"][SPANDATA.DB_SYSTEM] == "redis"
        assert span["data"][SPANDATA.DB_DRIVER_NAME] == "redis-py"
        assert span["data"][SPANDATA.DB_NAME] == "1"
        assert span["data"][SPANDATA.SERVER_ADDRESS] == "localhost"
        assert span["data"][SPANDATA.SERVER_PORT] == 63791


@pytest.mark.parametrize("span_streaming", [True, False])
def test_span_origin(sentry_init, capture_events, capture_items, span_streaming):
    sentry_init(
        integrations=[RedisIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    connection = FakeStrictRedis()

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            # default case
            connection.set("somekey", "somevalue")

            # pipeline
            pipeline = connection.pipeline(transaction=False)
            pipeline.get("somekey")
            pipeline.set("anotherkey", 1)
            pipeline.execute()
        sentry_sdk.flush()

        assert len(items) == 3
        set_span, pipeline_span, parent_span = [item.payload for item in items]

        assert parent_span["name"] == "custom parent"
        assert parent_span["attributes"]["sentry.origin"] == "manual"
        assert set_span["attributes"]["sentry.origin"] == "auto.db.redis"
        assert (
            set_span["attributes"][SPANDATA.DB_QUERY_TEXT] == "SET 'somekey' [Filtered]"
        )
        assert pipeline_span["attributes"]["sentry.origin"] == "auto.db.redis"
    else:
        events = capture_events()
        with start_transaction(name="custom_transaction"):
            # default case
            connection.set("somekey", "somevalue")

            # pipeline
            pipeline = connection.pipeline(transaction=False)
            pipeline.get("somekey")
            pipeline.set("anotherkey", 1)
            pipeline.execute()

        (event,) = events

        assert event["contexts"]["trace"]["origin"] == "manual"

        for span in event["spans"]:
            assert span["origin"] == "auto.db.redis"



from contextlib import asynccontextmanager
from functools import wraps

from sentry_sdk.integrations import DidNotEnable

from ..spans import (
    ai_client_span,
    update_ai_client_span,
)

try:
    from pydantic_ai._agent_graph import ModelRequestNode  # type: ignore
except ImportError:
    raise DidNotEnable("pydantic-ai not installed")

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable


def _extract_span_data(node: "Any", ctx: "Any") -> "tuple[list[Any], Any, Any]":
    """Extract common data needed for creating chat spans.

    Returns:
        Tuple of (messages, model, model_settings)
    """
    # Extract model and settings from context
    model = None
    model_settings = None
    if hasattr(ctx, "deps"):
        model = getattr(ctx.deps, "model", None)
        model_settings = getattr(ctx.deps, "model_settings", None)

    # Build full message list: history + current request
    messages = []
    if hasattr(ctx, "state") and hasattr(ctx.state, "message_history"):
        messages.extend(ctx.state.message_history)

    current_request = getattr(node, "request", None)
    if current_request:
        messages.append(current_request)

    return messages, model, model_settings


def _patch_graph_nodes() -> None:
    """
    Patches the graph node execution to create appropriate spans.

    ModelRequestNode -> Creates ai_client span for model requests
    CallToolsNode -> Handles tool calls (spans created in tool patching)
    """

    # Patch ModelRequestNode to create ai_client spans
    original_model_request_run = ModelRequestNode.run

    @wraps(original_model_request_run)
    async def wrapped_model_request_run(self: "Any", ctx: "Any") -> "Any":
        messages, model, model_settings = _extract_span_data(self, ctx)

        with ai_client_span(messages, None, model, model_settings) as span:
            result = await original_model_request_run(self, ctx)

            # Extract response from result if available
            model_response = None
            if hasattr(result, "model_response"):
                model_response = result.model_response

            update_ai_client_span(span, model_response)
            return result

    ModelRequestNode.run = wrapped_model_request_run

    # Patch ModelRequestNode.stream for streaming requests
    original_model_request_stream = ModelRequestNode.stream

    def create_wrapped_stream(
        original_stream_method: "Callable[..., Any]",
    ) -> "Callable[..., Any]":
        """Create a wrapper for ModelRequestNode.stream that creates chat spans."""

        @asynccontextmanager
        @wraps(original_stream_method)
        async def wrapped_model_request_stream(self: "Any", ctx: "Any") -> "Any":
            messages, model, model_settings = _extract_span_data(self, ctx)

            # Create chat span for streaming request
            with ai_client_span(messages, None, model, model_settings) as span:
                # Call the original stream method
                async with original_stream_method(self, ctx) as stream:
                    yield stream

                # After streaming completes, update span with response data
                # The ModelRequestNode stores the final response in _result
                model_response = None
                if hasattr(self, "_result") and self._result is not None:
                    # _result is a NextNode containing the model_response
                    if hasattr(self._result, "model_response"):
                        model_response = self._result.model_response

                update_ai_client_span(span, model_response)

        return wrapped_model_request_stream

    ModelRequestNode.stream = create_wrapped_stream(original_model_request_stream)



import random
from collections import Counter
from unittest import mock

import pytest

import sentry_sdk
from sentry_sdk import capture_exception, start_span, start_transaction
from sentry_sdk.tracing_utils import Baggage
from sentry_sdk.utils import logger


def test_sampling_decided_only_for_transactions(sentry_init, capture_events):
    sentry_init(traces_sample_rate=0.5)

    with start_transaction(name="hi") as transaction:
        assert transaction.sampled is not None

        with start_span() as span:
            assert span.sampled == transaction.sampled

    with start_span() as span:
        assert span.sampled is None


def test_sampling_decided_only_for_segments(sentry_init, capture_events):
    sentry_init(
        traces_sample_rate=0.5,
        trace_lifecycle="stream",
    )

    with sentry_sdk.traces.start_span(name="hi") as segment:
        assert segment.sampled is not None

        with sentry_sdk.traces.start_span(name="hey") as span:
            assert span.sampled == segment.sampled


@pytest.mark.parametrize("sampled", [True, False])
def test_nested_transaction_sampling_override(sentry_init, sampled):
    sentry_init(traces_sample_rate=1.0)

    with start_transaction(name="outer", sampled=sampled) as outer_transaction:
        assert outer_transaction.sampled is sampled
        with start_transaction(
            name="inner", sampled=(not sampled)
        ) as inner_transaction:
            assert inner_transaction.sampled is not sampled
        assert outer_transaction.sampled is sampled


def test_no_double_sampling(sentry_init, capture_events):
    # Transactions should not be subject to the global/error sample rate.
    # Only the traces_sample_rate should apply.
    sentry_init(traces_sample_rate=1.0, sample_rate=0.0)
    events = capture_events()

    with start_transaction(name="/"):
        pass

    assert len(events) == 1


def test_no_double_sampling_span_streaming(sentry_init, capture_items):
    # Segments should not be subject to the global/error sample rate.
    # Only the traces_sample_rate should apply.
    sentry_init(
        traces_sample_rate=1.0,
        sample_rate=0.0,
        trace_lifecycle="stream",
    )
    items = capture_items()

    with sentry_sdk.traces.start_span(name="/"):
        pass

    sentry_sdk.flush()

    assert len(items) == 1


@pytest.mark.parametrize("sampling_decision", [True, False])
def test_get_transaction_and_span_from_scope_regardless_of_sampling_decision(
    sentry_init, sampling_decision
):
    sentry_init(traces_sample_rate=1.0)

    with start_transaction(name="/", sampled=sampling_decision):
        with start_span(op="child-span"):
            with start_span(op="child-child-span"):
                scope = sentry_sdk.get_current_scope()
                assert scope.span.op == "child-child-span"
                assert scope.transaction.name == "/"


@pytest.mark.parametrize(
    "traces_sample_rate,expected_decision",
    [(0.0, False), (0.25, False), (0.75, True), (1.00, True)],
)
def test_uses_traces_sample_rate_correctly(
    sentry_init,
    traces_sample_rate,
    expected_decision,
):
    sentry_init(traces_sample_rate=traces_sample_rate)

    baggage = Baggage(sentry_items={"sample_rand": "0.500000"})
    transaction = start_transaction(name="dogpark", baggage=baggage)
    assert transaction.sampled is expected_decision


@pytest.mark.parametrize(
    "traces_sample_rate,expected_decision",
    [(0.0, False), (0.25, False), (0.75, True), (1.00, True)],
)
def test_uses_traces_sample_rate_correctly_span_streaming(
    sentry_init,
    traces_sample_rate,
    expected_decision,
):
    sentry_init(
        traces_sample_rate=traces_sample_rate,
        _experiments={
            "trace_lifecycle": "stream",
        },
    )

    sentry_sdk.traces.continue_trace(
        {
            "sentry-trace": "0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331",
            "baggage": "sentry-sample_rand=0.500000",
        }
    )

    with sentry_sdk.traces.start_span(name="dogpark") as span:
        assert span.sampled is expected_decision


@pytest.mark.parametrize(
    "traces_sampler_return_value,expected_decision",
    [(0.0, False), (0.25, False), (0.75, True), (1.00, True)],
)
def test_uses_traces_sampler_return_value_correctly(
    sentry_init,
    traces_sampler_return_value,
    expected_decision,
):
    sentry_init(traces_sampler=mock.Mock(return_value=traces_sampler_return_value))

    baggage = Baggage(sentry_items={"sample_rand": "0.500000"})
    transaction = start_transaction(name="dogpark", baggage=baggage)
    assert transaction.sampled is expected_decision


@pytest.mark.parametrize(
    "traces_sampler_return_value,expected_decision",
    [(0.0, False), (0.25, False), (0.75, True), (1.00, True)],
)
def test_uses_traces_sampler_return_value_correctly_span_streaming(
    sentry_init,
    traces_sampler_return_value,
    expected_decision,
):
    sentry_init(
        traces_sampler=mock.Mock(return_value=traces_sampler_return_value),
        _experiments={
            "trace_lifecycle": "stream",
        },
    )

    sentry_sdk.traces.continue_trace(
        {
            "sentry-trace": "0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331",
            "baggage": "sentry-sample_rand=0.500000",
        }
    )

    with sentry_sdk.traces.start_span(name="dogpark") as span:
        assert span.sampled is expected_decision


@pytest.mark.parametrize("traces_sampler_return_value", [True, False])
def test_tolerates_traces_sampler_returning_a_boolean(
    sentry_init, traces_sampler_return_value
):
    sentry_init(traces_sampler=mock.Mock(return_value=traces_sampler_return_value))

    transaction = start_transaction(name="dogpark")
    assert transaction.sampled is traces_sampler_return_value


@pytest.mark.parametrize("traces_sampler_return_value", [True, False])
def test_tolerates_traces_sampler_returning_a_boolean_span_streaming(
    sentry_init, traces_sampler_return_value
):
    sentry_init(
        traces_sampler=mock.Mock(return_value=traces_sampler_return_value),
        trace_lifecycle="stream",
    )

    with sentry_sdk.traces.start_span(name="dogpark") as span:
        assert span.sampled is traces_sampler_return_value


@pytest.mark.parametrize(
    "traces_sample_rate,expected_decision",
    [(0.0, False), (0.25, False), (0.75, True), (1.00, True)],
)
def test_traces_sampler_raising_falls_back_to_traces_sample_rate(
    sentry_init,
    traces_sample_rate,
    expected_decision,
):
    sentry_init(
        traces_sampler=mock.Mock(side_effect=ValueError("boom")),
        traces_sample_rate=traces_sample_rate,
    )

    baggage = Baggage(sentry_items={"sample_rand": "0.500000"})
    transaction = start_transaction(name="dogpark", baggage=baggage)
    assert transaction.sampled is expected_decision


@pytest.mark.parametrize("parent_sampling_decision", [True, False])
def test_traces_sampler_raising_falls_back_to_parent_sampling_decision(
    sentry_init, parent_sampling_decision
):
    # set traces_sample_rate to produce the opposite of the parent decision,
    # to prove the parent's decision takes precedence in the fallback
    sentry_init(
        traces_sampler=mock.Mock(side_effect=ValueError("boom")),
        traces_sample_rate=0.0 if parent_sampling_decision else 1.0,
    )

    baggage = Baggage(sentry_items={"sample_rand": "0.500000"})
    transaction = start_transaction(
        name="dogpark", baggage=baggage, parent_sampled=parent_sampling_decision
    )
    assert transaction.sampled is parent_sampling_decision


@pytest.mark.parametrize(
    "traces_sample_rate,expected_decision",
    [(0.0, False), (0.25, False), (0.75, True), (1.00, True)],
)
def test_traces_sampler_raising_falls_back_to_traces_sample_rate_span_streaming(
    sentry_init,
    traces_sample_rate,
    expected_decision,
):
    sentry_init(
        traces_sampler=mock.Mock(side_effect=ValueError("boom")),
        traces_sample_rate=traces_sample_rate,
        trace_lifecycle="stream",
    )

    sentry_sdk.traces.continue_trace(
        {
            "sentry-trace": "0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331",
            "baggage": "sentry-sample_rand=0.500000",
        }
    )

    with sentry_sdk.traces.start_span(name="dogpark") as span:
        assert span.sampled is expected_decision


@pytest.mark.parametrize(
    "traces_sample_rate,expected_decision",
    [(0.0, False), (0.25, False), (0.75, True), (1.00, True)],
)
def test_traces_sampler_raising_no_incoming_trace_falls_back_to_traces_sample_rate_span_streaming(
    sentry_init,
    traces_sample_rate,
    expected_decision,
    monkeypatch,
):
    sentry_init(
        traces_sampler=mock.Mock(side_effect=ValueError("boom")),
        traces_sample_rate=traces_sample_rate,
        trace_lifecycle="stream",
    )

    # no continue_trace, so no propagated sample_rand; make it deterministic
    monkeypatch.setattr(
        "sentry_sdk.tracing_utils._generate_sample_rand", lambda *a, **kw: 0.5
    )

    with sentry_sdk.traces.start_span(name="dogpark") as span:
        assert span.sampled is expected_decision


def test_traces_sampler_raising_no_incoming_trace_and_no_traces_sample_rate(
    sentry_init,
):
    sentry_init(
        traces_sampler=mock.Mock(side_effect=ValueError("boom")),
    )

    transaction = start_transaction(name="dogpark")
    assert transaction.sampled is False


def test_traces_sampler_raising_no_incoming_trace_and_no_traces_sample_rate_span_streaming(
    sentry_init,
):
    sentry_init(
        traces_sampler=mock.Mock(side_effect=ValueError("boom")),
        trace_lifecycle="stream",
    )

    with sentry_sdk.traces.start_span(name="dogpark") as span:
        assert span.sampled is False


@pytest.mark.parametrize("sampling_decision", [True, False])
def test_only_captures_transaction_when_sampled_is_true(
    sentry_init, sampling_decision, capture_events
):
    sentry_init(traces_sampler=mock.Mock(return_value=sampling_decision))
    events = capture_events()

    transaction = start_transaction(name="dogpark")
    transaction.finish()

    assert len(events) == (1 if sampling_decision else 0)


@pytest.mark.parametrize("sampling_decision", [True, False])
def test_only_captures_segment_when_sampled_is_true_span_streaming(
    sentry_init, sampling_decision, capture_items
):
    sentry_init(
        traces_sampler=mock.Mock(
            return_value=sampling_decision,
        ),
        _experiments={
            "trace_lifecycle": "stream",
        },
    )
    items = capture_items()

    span = sentry_sdk.traces.start_span(name="dogpark")
    span.end()

    sentry_sdk.flush()

    assert len(items) == (1 if sampling_decision else 0)


@pytest.mark.parametrize(
    "traces_sample_rate,traces_sampler_return_value", [(0, True), (1, False)]
)
def test_prefers_traces_sampler_to_traces_sample_rate(
    sentry_init,
    traces_sample_rate,
    traces_sampler_return_value,
):
    # make traces_sample_rate imply the opposite of traces_sampler, to prove
    # that traces_sampler takes precedence
    traces_sampler = mock.Mock(return_value=traces_sampler_return_value)
    sentry_init(
        traces_sample_rate=traces_sample_rate,
        traces_sampler=traces_sampler,
    )

    transaction = start_transaction(name="dogpark")
    assert traces_sampler.called is True
    assert transaction.sampled is traces_sampler_return_value


@pytest.mark.parametrize(
    "traces_sample_rate,traces_sampler_return_value", [(0, True), (1, False)]
)
def test_prefers_traces_sampler_to_traces_sample_rate_span_streaming(
    sentry_init,
    traces_sample_rate,
    traces_sampler_return_value,
):
    # make traces_sample_rate imply the opposite of traces_sampler, to prove
    # that traces_sampler takes precedence
    traces_sampler = mock.Mock(return_value=traces_sampler_return_value)
    sentry_init(
        traces_sample_rate=traces_sample_rate,
        traces_sampler=traces_sampler,
        _experiments={
            "trace_lifecycle": "stream",
        },
    )

    span = sentry_sdk.traces.start_span(name="dogpark")
    assert traces_sampler.called is True
    assert span.sampled is traces_sampler_return_value


@pytest.mark.parametrize("parent_sampling_decision", [True, False])
def test_ignores_inherited_sample_decision_when_traces_sampler_defined(
    sentry_init, parent_sampling_decision
):
    # make traces_sampler pick the opposite of the inherited decision, to prove
    # that traces_sampler takes precedence
    traces_sampler = mock.Mock(return_value=not parent_sampling_decision)
    sentry_init(traces_sampler=traces_sampler)

    transaction = start_transaction(
        name="dogpark", parent_sampled=parent_sampling_decision
    )
    assert transaction.sampled is not parent_sampling_decision


@pytest.mark.parametrize("parent_sampling_decision", ["1", "0"])
def test_ignores_inherited_sample_decision_when_traces_sampler_defined_span_streaming(
    sentry_init, parent_sampling_decision
):
    # make traces_sampler pick the opposite of the inherited decision, to prove
    # that traces_sampler takes precedence
    traces_sampler = mock.Mock(return_value=not bool(int(parent_sampling_decision)))
    sentry_init(
        traces_sampler=traces_sampler,
        trace_lifecycle="stream",
    )

    sentry_sdk.traces.continue_trace(
        {
            "sentry-trace": f"0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-{parent_sampling_decision}"
        }
    )
    span = sentry_sdk.traces.start_span(name="dogpark")
    assert span.sampled is not bool(int(parent_sampling_decision))


@pytest.mark.parametrize("explicit_decision", [True, False])
def test_traces_sampler_doesnt_overwrite_explicitly_passed_sampling_decision(
    sentry_init, explicit_decision
):
    # make traces_sampler pick the opposite of the explicit decision, to prove
    # that the explicit decision takes precedence
    traces_sampler = mock.Mock(return_value=not explicit_decision)
    sentry_init(traces_sampler=traces_sampler)

    transaction = start_transaction(name="dogpark", sampled=explicit_decision)
    assert transaction.sampled is explicit_decision


@pytest.mark.parametrize("parent_sampling_decision", [True, False])
def test_inherits_parent_sampling_decision_when_traces_sampler_undefined(
    sentry_init, parent_sampling_decision
):
    # make sure the parent sampling decision is the opposite of what
    # traces_sample_rate would produce, to prove the inheritance takes
    # precedence
    sentry_init(traces_sample_rate=0.5)
    mock_random_value = 0.25 if parent_sampling_decision is False else 0.75

    with mock.patch.object(random, "random", return_value=mock_random_value):
        transaction = start_transaction(
            name="dogpark", parent_sampled=parent_sampling_decision
        )
        assert transaction.sampled is parent_sampling_decision


@pytest.mark.parametrize("parent_sampling_decision", ["1", "0"])
def test_inherits_parent_sampling_decision_when_traces_sampler_undefined_span_streaming(
    sentry_init, parent_sampling_decision
):
    sentry_init(
        traces_sample_rate=0.5,
        trace_lifecycle="stream",
    )

    sentry_sdk.traces.continue_trace(
        {
            "sentry-trace": f"0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-{parent_sampling_decision}"
        }
    )

    # make sure the parent sampling decision is the opposite of what
    # traces_sample_rate would produce, to prove the inheritance takes
    # precedence
    mock_random_value = 0.25 if parent_sampling_decision == "0" else 0.75
    with mock.patch.object(random, "random", return_value=mock_random_value):
        span = sentry_sdk.traces.start_span(name="dogpark")
        assert span.sampled is bool(int(parent_sampling_decision))


@pytest.mark.parametrize("parent_sampling_decision", [True, False])
def test_passes_parent_sampling_decision_in_sampling_context(
    sentry_init, parent_sampling_decision
):
    sentry_init(traces_sample_rate=1.0)

    sentry_trace_header = (
        "12312012123120121231201212312012-1121201211212012-{sampled}".format(
            sampled=int(parent_sampling_decision)
        )
    )

    transaction = sentry_sdk.continue_trace(
        {"sentry-trace": sentry_trace_header},
        name="dogpark",
    )

    def mock_set_initial_sampling_decision(_, sampling_context):
        assert "parent_sampled" in sampling_context
        assert sampling_context["parent_sampled"] is parent_sampling_decision

    with mock.patch(
        "sentry_sdk.tracing.Transaction._set_initial_sampling_decision",
        mock_set_initial_sampling_decision,
    ):
        start_transaction(transaction=transaction)


def test_passes_custom_sampling_context_from_start_transaction_to_traces_sampler(
    sentry_init,
    DictionaryContaining,  # noqa: N803
):
    traces_sampler = mock.Mock()
    sentry_init(traces_sampler=traces_sampler)

    start_transaction(custom_sampling_context={"dogs": "yes", "cats": "maybe"})

    traces_sampler.assert_any_call(
        DictionaryContaining({"dogs": "yes", "cats": "maybe"})
    )


def test_custom_sampling_context(sentry_init):
    class MyClass: ...

    my_class = MyClass()

    def traces_sampler(sampling_context):
        assert "class" in sampling_context
        assert "string" in sampling_context
        assert sampling_context["class"] == my_class
        assert sampling_context["string"] == "my string"
        return 1.0

    sentry_init(
        traces_sampler=traces_sampler,
        trace_lifecycle="stream",
    )

    sentry_sdk.get_current_scope().set_custom_sampling_context(
        {
            "class": my_class,
            "string": "my string",
        }
    )

    with sentry_sdk.traces.start_span(name="span"):
        ...


def test_custom_sampling_context_update_to_context_value_persists(sentry_init):
    def traces_sampler(sampling_context):
        if sampling_context["transaction_context"]["data"]["first"] is True:
            assert sampling_context["custom_value"] == 1
        else:
            assert sampling_context["custom_value"] == 2
        return 1.0

    sentry_init(
        traces_sampler=traces_sampler,
        trace_lifecycle="stream",
    )

    sentry_sdk.traces.new_trace()

    sentry_sdk.get_current_scope().set_custom_sampling_context({"custom_value": 1})

    with sentry_sdk.traces.start_span(name="span", attributes={"first": True}):
        ...

    sentry_sdk.traces.new_trace()

    sentry_sdk.get_current_scope().set_custom_sampling_context({"custom_value": 2})

    with sentry_sdk.traces.start_span(name="span", attributes={"first": False}):
        ...


def test_sample_rate_affects_errors(sentry_init, capture_events):
    sentry_init(sample_rate=0)
    events = capture_events()

    try:
        1 / 0
    except Exception:
        capture_exception()

    assert len(events) == 0


@pytest.mark.parametrize(
    "traces_sampler_return_value",
    [
        "dogs are great",  # wrong type
        (0, 1),  # wrong type
        {"Maisey": "Charllie"},  # wrong type
        [True, True],  # wrong type
        {0.2012},  # wrong type
        float("NaN"),  # wrong type
        None,  # wrong type
        -1.121,  # wrong value
        1.231,  # wrong value
    ],
)
def test_warns_and_sets_sampled_to_false_on_invalid_traces_sampler_return_value(
    sentry_init,
    traces_sampler_return_value,
    StringContaining,  # noqa: N803
):
    sentry_init(traces_sampler=mock.Mock(return_value=traces_sampler_return_value))

    with mock.patch.object(logger, "warning", mock.Mock()):
        transaction = start_transaction(name="dogpark")
        logger.warning.assert_any_call(StringContaining("Given sample rate is invalid"))
        assert transaction.sampled is False


@pytest.mark.parametrize(
    "traces_sampler_return_value",
    [
        "dogs are great",  # wrong type
        (0, 1),  # wrong type
        {"Maisey": "Charllie"},  # wrong type
        [True, True],  # wrong type
        {0.2012},  # wrong type
        float("NaN"),  # wrong type
        None,  # wrong type
        -1.121,  # wrong value
        1.231,  # wrong value
    ],
)
def test_warns_and_sets_sampled_to_false_on_invalid_traces_sampler_return_value_span_streaming(
    sentry_init,
    traces_sampler_return_value,
    StringContaining,  # noqa: N803
):
    sentry_init(
        traces_sampler=mock.Mock(return_value=traces_sampler_return_value),
        trace_lifecycle="stream",
    )

    with mock.patch.object(logger, "warning", mock.Mock()):
        span = sentry_sdk.traces.start_span(name="dogpark")
        logger.warning.assert_any_call(StringContaining("Given sample rate is invalid"))
        assert span.sampled is False


@pytest.mark.parametrize(
    "traces_sample_rate,sampled_output,expected_record_lost_event_calls",
    [
        (None, False, []),
        (
            0.0,
            False,
            [("sample_rate", "transaction", None, 1), ("sample_rate", "span", None, 1)],
        ),
        (1.0, True, []),
    ],
)
def test_records_lost_event_only_if_traces_sample_rate_enabled(
    sentry_init,
    capture_record_lost_event_calls,
    traces_sample_rate,
    sampled_output,
    expected_record_lost_event_calls,
):
    sentry_init(traces_sample_rate=traces_sample_rate)
    record_lost_event_calls = capture_record_lost_event_calls()

    transaction = start_transaction(name="dogpark")
    assert transaction.sampled is sampled_output
    transaction.finish()

    # Use Counter because order of calls does not matter
    assert Counter(record_lost_event_calls) == Counter(expected_record_lost_event_calls)


@pytest.mark.parametrize(
    "traces_sampler,sampled_output,expected_record_lost_event_calls",
    [
        (None, False, []),
        (
            lambda _x: 0.0,
            False,
            [("sample_rate", "transaction", None, 1), ("sample_rate", "span", None, 1)],
        ),
        (lambda _x: 1.0, True, []),
    ],
)
def test_records_lost_event_only_if_traces_sampler_enabled(
    sentry_init,
    capture_record_lost_event_calls,
    traces_sampler,
    sampled_output,
    expected_record_lost_event_calls,
):
    sentry_init(traces_sampler=traces_sampler)
    record_lost_event_calls = capture_record_lost_event_calls()

    transaction = start_transaction(name="dogpark")
    assert transaction.sampled is sampled_output
    transaction.finish()

    # Use Counter because order of calls does not matter
    assert Counter(record_lost_event_calls) == Counter(expected_record_lost_event_calls)


def test_unsampled_spans_produce_client_report_if_traces_sample_rate_defined(
    sentry_init, capture_items, capture_record_lost_event_calls
):
    sentry_init(
        traces_sample_rate=0.0,
        trace_lifecycle="stream",
    )

    items = capture_items("span")
    record_lost_event_calls = capture_record_lost_event_calls()

    with sentry_sdk.traces.start_span(name="segment"):
        with sentry_sdk.traces.start_span(name="child1"):
            pass
        with sentry_sdk.traces.start_span(name="child2"):
            pass

    sentry_sdk.get_client().flush()

    spans = [item.payload for item in items]
    assert not spans

    assert record_lost_event_calls == [
        ("sample_rate", "span", None, 1),
        ("sample_rate", "span", None, 1),
        ("sample_rate", "span", None, 1),
    ]


def test_unsampled_spans_produce_client_report_if_traces_sampler_defined(
    sentry_init, capture_items, capture_record_lost_event_calls
):
    sentry_init(
        traces_sampler=lambda _: 0.0,
        trace_lifecycle="stream",
    )

    items = capture_items("span")
    record_lost_event_calls = capture_record_lost_event_calls()

    with sentry_sdk.traces.start_span(name="segment"):
        with sentry_sdk.traces.start_span(name="child1"):
            pass
        with sentry_sdk.traces.start_span(name="child2"):
            pass

    sentry_sdk.get_client().flush()

    spans = [item.payload for item in items]
    assert not spans

    assert record_lost_event_calls == [
        ("sample_rate", "span", None, 1),
        ("sample_rate", "span", None, 1),
        ("sample_rate", "span", None, 1),
    ]


def test_no_client_reports_if_tracing_is_off(
    sentry_init, capture_items, capture_record_lost_event_calls
):
    sentry_init(
        traces_sample_rate=None,
        trace_lifecycle="stream",
    )

    items = capture_items("span")
    record_lost_event_calls = capture_record_lost_event_calls()

    with sentry_sdk.traces.start_span(name="segment"):
        with sentry_sdk.traces.start_span(name="child1"):
            pass
        with sentry_sdk.traces.start_span(name="child2"):
            pass

    sentry_sdk.get_client().flush()

    spans = [item.payload for item in items]
    assert not spans
    assert not record_lost_event_calls



import os

import sentry_sdk
from sentry_sdk.integrations.aws_lambda import AwsLambdaIntegration

sentry_sdk.init(
    dsn=os.environ.get("SENTRY_DSN"),
    traces_sample_rate=None,  # this is the default, just added for clarity
    integrations=[AwsLambdaIntegration()],
)


def handler(event, context):
    raise Exception("Oh!")



import uuid
from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.utils import logger

if TYPE_CHECKING:
    from typing import Optional

    from sentry_sdk._types import Event, MonitorConfig


def _create_check_in_event(
    monitor_slug: "Optional[str]" = None,
    check_in_id: "Optional[str]" = None,
    status: "Optional[str]" = None,
    duration_s: "Optional[float]" = None,
    monitor_config: "Optional[MonitorConfig]" = None,
) -> "Event":
    options = sentry_sdk.get_client().options
    check_in_id = check_in_id or uuid.uuid4().hex

    check_in: "Event" = {
        "type": "check_in",
        "monitor_slug": monitor_slug,
        "check_in_id": check_in_id,
        "status": status,
        "duration": duration_s,
        "environment": options.get("environment", None),
        "release": options.get("release", None),
    }

    if monitor_config:
        check_in["monitor_config"] = monitor_config

    return check_in


def capture_checkin(
    monitor_slug: "Optional[str]" = None,
    check_in_id: "Optional[str]" = None,
    status: "Optional[str]" = None,
    duration: "Optional[float]" = None,
    monitor_config: "Optional[MonitorConfig]" = None,
) -> str:
    check_in_event = _create_check_in_event(
        monitor_slug=monitor_slug,
        check_in_id=check_in_id,
        status=status,
        duration_s=duration,
        monitor_config=monitor_config,
    )

    sentry_sdk.capture_event(check_in_event)

    logger.debug(
        f"[Crons] Captured check-in ({check_in_event.get('check_in_id')}): {check_in_event.get('monitor_slug')} -> {check_in_event.get('status')}"
    )

    return check_in_event["check_in_id"]



import asyncio
from datetime import timedelta

import arq.worker
import pytest
from arq import cron
from arq.connections import ArqRedis
from arq.jobs import Job
from arq.utils import timestamp_ms
from fakeredis.aioredis import FakeRedis

import sentry_sdk
from sentry_sdk import get_client, start_transaction
from sentry_sdk.consts import SPANDATA
from sentry_sdk.integrations.arq import ArqIntegration


def async_partial(async_fn, *args, **kwargs):
    # asyncio.iscoroutinefunction (Used in the integration code) in Python < 3.8
    # does not detect async functions in functools.partial objects.
    # This partial implementation returns a coroutine instead.
    async def wrapped(ctx):
        return await async_fn(ctx, *args, **kwargs)

    return wrapped


@pytest.fixture(autouse=True)
def patch_fakeredis_info_command():
    from fakeredis._fakesocket import FakeSocket

    if not hasattr(FakeSocket, "info"):
        from fakeredis._commands import command
        from fakeredis._helpers import SimpleString

        @command((SimpleString,), name="info")
        def info(self, section):
            return section

        FakeSocket.info = info


@pytest.fixture
def init_arq(sentry_init):
    def inner(
        span_streaming,
        cls_functions=None,
        cls_cron_jobs=None,
        kw_functions=None,
        kw_cron_jobs=None,
        allow_abort_jobs_=False,
    ):
        cls_functions = cls_functions or []
        cls_cron_jobs = cls_cron_jobs or []

        kwargs = {}
        if kw_functions is not None:
            kwargs["functions"] = kw_functions
        if kw_cron_jobs is not None:
            kwargs["cron_jobs"] = kw_cron_jobs

        sentry_init(
            integrations=[ArqIntegration()],
            traces_sample_rate=1.0,
            send_default_pii=True,
            trace_lifecycle="stream" if span_streaming else "static",
        )

        server = FakeRedis()
        pool = ArqRedis(pool_or_conn=server.connection_pool)

        class WorkerSettings:
            functions = cls_functions
            cron_jobs = cls_cron_jobs
            redis_pool = pool
            allow_abort_jobs = allow_abort_jobs_

        if not WorkerSettings.functions:
            del WorkerSettings.functions
        if not WorkerSettings.cron_jobs:
            del WorkerSettings.cron_jobs

        worker = arq.worker.create_worker(WorkerSettings, **kwargs)

        return pool, worker

    return inner


@pytest.fixture
def init_arq_with_dict_settings(sentry_init):
    def inner(
        span_streaming,
        cls_functions=None,
        cls_cron_jobs=None,
        kw_functions=None,
        kw_cron_jobs=None,
        allow_abort_jobs_=False,
    ):
        cls_functions = cls_functions or []
        cls_cron_jobs = cls_cron_jobs or []

        kwargs = {}
        if kw_functions is not None:
            kwargs["functions"] = kw_functions
        if kw_cron_jobs is not None:
            kwargs["cron_jobs"] = kw_cron_jobs

        sentry_init(
            integrations=[ArqIntegration()],
            traces_sample_rate=1.0,
            send_default_pii=True,
            trace_lifecycle="stream" if span_streaming else "static",
        )

        server = FakeRedis()
        pool = ArqRedis(pool_or_conn=server.connection_pool)

        worker_settings = {
            "functions": cls_functions,
            "cron_jobs": cls_cron_jobs,
            "redis_pool": pool,
            "allow_abort_jobs": allow_abort_jobs_,
        }

        if not worker_settings["functions"]:
            del worker_settings["functions"]
        if not worker_settings["cron_jobs"]:
            del worker_settings["cron_jobs"]

        worker = arq.worker.create_worker(worker_settings, **kwargs)

        return pool, worker

    return inner


@pytest.fixture
def init_arq_with_kwarg_settings(sentry_init):
    """Test fixture that passes settings_cls as keyword argument only."""

    def inner(
        span_streaming,
        cls_functions=None,
        cls_cron_jobs=None,
        kw_functions=None,
        kw_cron_jobs=None,
        allow_abort_jobs_=False,
    ):
        cls_functions = cls_functions or []
        cls_cron_jobs = cls_cron_jobs or []

        kwargs = {}
        if kw_functions is not None:
            kwargs["functions"] = kw_functions
        if kw_cron_jobs is not None:
            kwargs["cron_jobs"] = kw_cron_jobs

        sentry_init(
            integrations=[ArqIntegration()],
            traces_sample_rate=1.0,
            send_default_pii=True,
            trace_lifecycle="stream" if span_streaming else "static",
        )

        server = FakeRedis()
        pool = ArqRedis(pool_or_conn=server.connection_pool)

        class WorkerSettings:
            functions = cls_functions
            cron_jobs = cls_cron_jobs
            redis_pool = pool
            allow_abort_jobs = allow_abort_jobs_

        if not WorkerSettings.functions:
            del WorkerSettings.functions
        if not WorkerSettings.cron_jobs:
            del WorkerSettings.cron_jobs

        # Pass settings_cls as keyword argument (not positional)
        worker = arq.worker.create_worker(settings_cls=WorkerSettings, **kwargs)

        return pool, worker

    return inner


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "init_arq_settings",
    ["init_arq", "init_arq_with_dict_settings", "init_arq_with_kwarg_settings"],
)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_job_result(init_arq_settings, request, span_streaming):
    async def increase(ctx, num):
        return num + 1

    init_fixture_method = request.getfixturevalue(init_arq_settings)

    increase.__qualname__ = increase.__name__

    pool, worker = init_fixture_method(span_streaming, [increase])

    job = await pool.enqueue_job("increase", 3)

    assert isinstance(job, Job)

    await worker.run_job(job.job_id, timestamp_ms())
    result = await job.result()
    job_result = await job.result_info()

    assert result == 4
    assert job_result.result == 4


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "init_arq_settings", ["init_arq", "init_arq_with_dict_settings"]
)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_job_retry(
    capture_events,
    capture_items,
    init_arq_settings,
    request,
    span_streaming,
):
    async def retry_job(ctx):
        if ctx["job_try"] < 2:
            raise arq.worker.Retry

    init_fixture_method = request.getfixturevalue(init_arq_settings)

    retry_job.__qualname__ = retry_job.__name__

    pool, worker = init_fixture_method(span_streaming, [retry_job])

    job = await pool.enqueue_job("retry_job")

    if span_streaming:
        items = capture_items("span")

        await worker.run_job(job.job_id, timestamp_ms())

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        # The retry re-enqueue happens without an active span, so no producer
        # (queue.submit.arq) span is created for it; only the consumer segment
        # is emitted. The consumer segment is preceded by the redis spans for
        # the re-enqueue, so it lands at index 2.
        assert spans[2]["attributes"]["sentry.op"] == "queue.task.arq"
        assert spans[2]["status"] == "ok"
        assert spans[2]["name"] == "retry_job"

        await worker.run_job(job.job_id, timestamp_ms())

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        assert spans[5]["attributes"]["sentry.op"] == "queue.task.arq"
        assert spans[5]["status"] == "ok"
        assert spans[5]["name"] == "retry_job"
    else:
        events = capture_events()

        await worker.run_job(job.job_id, timestamp_ms())

        event = events.pop(0)
        assert event["contexts"]["trace"]["status"] == "aborted"
        assert event["transaction"] == "retry_job"
        assert event["tags"]["arq_task_id"] == job.job_id
        assert event["extra"]["arq-job"]["retry"] == 1

        await worker.run_job(job.job_id, timestamp_ms())

        event = events.pop(0)
        assert event["contexts"]["trace"]["status"] == "ok"
        assert event["transaction"] == "retry_job"
        assert event["tags"]["arq_task_id"] == job.job_id
        assert event["extra"]["arq-job"]["retry"] == 2


@pytest.mark.parametrize(
    "source", [("cls_functions", "cls_cron_jobs"), ("kw_functions", "kw_cron_jobs")]
)
@pytest.mark.parametrize("job_fails", [True, False], ids=["error", "success"])
@pytest.mark.parametrize(
    "init_arq_settings", ["init_arq", "init_arq_with_dict_settings"]
)
@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_job_transaction(
    capture_events,
    capture_items,
    init_arq_settings,
    source,
    job_fails,
    request,
    span_streaming,
):
    async def division(_, a, b=0):
        return a / b

    init_fixture_method = request.getfixturevalue(init_arq_settings)

    division.__qualname__ = division.__name__

    cron_func = async_partial(division, a=1, b=int(not job_fails))
    cron_func.__qualname__ = division.__name__

    cron_job = cron(cron_func, minute=0, run_at_startup=True)

    functions_key, cron_jobs_key = source
    pool, worker = init_fixture_method(
        span_streaming, **{functions_key: [division], cron_jobs_key: [cron_job]}
    )

    job = await pool.enqueue_job("division", 1, b=int(not job_fails))

    if span_streaming:
        items = capture_items("event", "span")

        await worker.run_job(job.job_id, timestamp_ms())

        loop = asyncio.get_event_loop()
        task = loop.create_task(worker.async_run())
        await asyncio.sleep(1)

        task.cancel()

        await worker.close()

        events = [item.payload for item in items if item.type == "event"]
        if job_fails:
            error_func_event = events.pop(0)
            error_cron_event = events.pop(0)

            assert (
                error_func_event["exception"]["values"][0]["type"]
                == "ZeroDivisionError"
            )
            assert (
                error_func_event["exception"]["values"][0]["mechanism"]["type"] == "arq"
            )

            func_extra = error_func_event["extra"]["arq-job"]
            assert func_extra["task"] == "division"

            assert (
                error_cron_event["exception"]["values"][0]["type"]
                == "ZeroDivisionError"
            )
            assert (
                error_cron_event["exception"]["values"][0]["mechanism"]["type"] == "arq"
            )

            cron_extra = error_cron_event["extra"]["arq-job"]
            assert cron_extra["task"] == "cron:division"

        sentry_sdk.flush()
        spans = [item.payload for item in items if item.type == "span"]

        task_spans = [
            span
            for span in spans
            if span["attributes"].get("sentry.op") == "queue.task.arq"
        ]

        division_span = next(span for span in task_spans if span["name"] == "division")
        assert division_span["attributes"]["sentry.segment.name.source"] == "task"
        assert (
            division_span["attributes"][SPANDATA.MESSAGING_DESTINATION_NAME]
            == worker.queue_name
        )

        assert any(span["name"] == "cron:division" for span in task_spans)
    else:
        events = capture_events()

        await worker.run_job(job.job_id, timestamp_ms())

        loop = asyncio.get_event_loop()
        task = loop.create_task(worker.async_run())
        await asyncio.sleep(1)

        task.cancel()

        await worker.close()

        if job_fails:
            error_func_event = events.pop(0)
            error_cron_event = events.pop(1)

            assert (
                error_func_event["exception"]["values"][0]["type"]
                == "ZeroDivisionError"
            )
            assert (
                error_func_event["exception"]["values"][0]["mechanism"]["type"] == "arq"
            )

            func_extra = error_func_event["extra"]["arq-job"]
            assert func_extra["task"] == "division"

            assert (
                error_cron_event["exception"]["values"][0]["type"]
                == "ZeroDivisionError"
            )
            assert (
                error_cron_event["exception"]["values"][0]["mechanism"]["type"] == "arq"
            )

            cron_extra = error_cron_event["extra"]["arq-job"]
            assert cron_extra["task"] == "cron:division"

        [func_event, cron_event] = events

        assert func_event["type"] == "transaction"
        assert func_event["transaction"] == "division"
        assert func_event["transaction_info"] == {"source": "task"}
        assert (
            func_event["contexts"]["trace"]["data"][SPANDATA.MESSAGING_DESTINATION_NAME]
            == worker.queue_name
        )

        assert "arq_task_id" in func_event["tags"]
        assert "arq_task_retry" in func_event["tags"]

        func_extra = func_event["extra"]["arq-job"]

        assert func_extra["task"] == "division"
        assert func_extra["kwargs"] == {"b": int(not job_fails)}
        assert func_extra["retry"] == 1

        assert cron_event["type"] == "transaction"
        assert cron_event["transaction"] == "cron:division"
        assert cron_event["transaction_info"] == {"source": "task"}

        assert "arq_task_id" in cron_event["tags"]
        assert "arq_task_retry" in cron_event["tags"]

        cron_extra = cron_event["extra"]["arq-job"]

        assert cron_extra["task"] == "cron:division"
        assert cron_extra["kwargs"] == {}
        assert cron_extra["retry"] == 1


@pytest.mark.parametrize("source", ["cls_functions", "kw_functions"])
@pytest.mark.parametrize(
    "init_arq_settings", ["init_arq", "init_arq_with_dict_settings"]
)
@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_enqueue_job(
    capture_events,
    capture_items,
    init_arq_settings,
    source,
    request,
    span_streaming,
):
    async def dummy_job(_):
        pass

    init_fixture_method = request.getfixturevalue(init_arq_settings)

    pool, _ = init_fixture_method(span_streaming, **{source: [dummy_job]})

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent") as span:
            await pool.enqueue_job("dummy_job")

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        assert spans[2]["is_segment"] is True
        assert spans[2]["trace_id"] == span.trace_id
        assert spans[2]["span_id"] == span.span_id

        assert spans[1]["attributes"]["sentry.op"] == "queue.submit.arq"
        assert spans[1]["name"] == "dummy_job"
    else:
        events = capture_events()

        with start_transaction() as transaction:
            await pool.enqueue_job("dummy_job")

        (event,) = events

        assert event["contexts"]["trace"]["trace_id"] == transaction.trace_id
        assert event["contexts"]["trace"]["span_id"] == transaction.span_id

        assert len(event["spans"])
        assert event["spans"][0]["op"] == "queue.submit.arq"
        assert event["spans"][0]["description"] == "dummy_job"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "init_arq_settings", ["init_arq", "init_arq_with_dict_settings"]
)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_execute_job_without_integration(
    init_arq_settings, request, span_streaming
):
    async def dummy_job(_ctx):
        pass

    init_fixture_method = request.getfixturevalue(init_arq_settings)

    dummy_job.__qualname__ = dummy_job.__name__

    pool, worker = init_fixture_method(span_streaming, [dummy_job])
    # remove the integration to trigger the edge case
    get_client().integrations.pop("arq")

    job = await pool.enqueue_job("dummy_job")

    await worker.run_job(job.job_id, timestamp_ms())

    assert await job.result() is None


@pytest.mark.parametrize("source", ["cls_functions", "kw_functions"])
@pytest.mark.parametrize(
    "init_arq_settings", ["init_arq", "init_arq_with_dict_settings"]
)
@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_span_origin_producer(
    capture_events,
    capture_items,
    init_arq_settings,
    source,
    request,
    span_streaming,
):
    async def dummy_job(_):
        pass

    init_fixture_method = request.getfixturevalue(init_arq_settings)

    pool, _ = init_fixture_method(span_streaming, **{source: [dummy_job]})

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            await pool.enqueue_job("dummy_job")

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        assert spans[2]["attributes"]["sentry.origin"] == "manual"
        assert spans[1]["attributes"]["sentry.origin"] == "auto.queue.arq"
    else:
        events = capture_events()

        with start_transaction():
            await pool.enqueue_job("dummy_job")

        (event,) = events
        assert event["contexts"]["trace"]["origin"] == "manual"
        assert event["spans"][0]["origin"] == "auto.queue.arq"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "init_arq_settings", ["init_arq", "init_arq_with_dict_settings"]
)
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_span_origin_consumer(
    capture_events,
    capture_items,
    init_arq_settings,
    request,
    span_streaming,
):
    async def job(ctx):
        pass

    init_fixture_method = request.getfixturevalue(init_arq_settings)

    job.__qualname__ = job.__name__

    pool, worker = init_fixture_method(span_streaming, [job])

    if span_streaming:
        job = await pool.enqueue_job("job")

        items = capture_items("span")

        await worker.run_job(job.job_id, timestamp_ms())

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        # No producer (queue.submit.arq) span is created for the re-enqueue
        # triggered by the retry, since it happens without an active span, so
        # the consumer segment lands at index 2.
        assert spans[2]["attributes"]["sentry.op"] == "queue.task.arq"
        assert spans[2]["attributes"]["sentry.origin"] == "auto.queue.arq"
        assert spans[1]["attributes"]["sentry.origin"] == "auto.db.redis"
        assert spans[0]["attributes"]["sentry.origin"] == "auto.db.redis"
    else:
        job = await pool.enqueue_job("job")

        events = capture_events()

        await worker.run_job(job.job_id, timestamp_ms())

        (event,) = events

        assert event["contexts"]["trace"]["origin"] == "auto.queue.arq"
        assert event["spans"][0]["origin"] == "auto.db.redis"
        assert event["spans"][1]["origin"] == "auto.db.redis"


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_job_concurrency(
    capture_events,
    capture_items,
    init_arq,
    span_streaming,
):
    """
    10 - division starts
    70 - sleepy starts
    110 - division raises error
    120 - sleepy finishes

    """

    async def sleepy(_):
        await asyncio.sleep(0.05)

    async def division(_):
        await asyncio.sleep(0.1)
        return 1 / 0

    sleepy.__qualname__ = sleepy.__name__
    division.__qualname__ = division.__name__

    pool, worker = init_arq(span_streaming, [sleepy, division])

    await pool.enqueue_job(
        "division", _job_id="123", _defer_by=timedelta(milliseconds=10)
    )
    await pool.enqueue_job(
        "sleepy", _job_id="456", _defer_by=timedelta(milliseconds=70)
    )

    if span_streaming:
        items = capture_items("event")

        loop = asyncio.get_event_loop()
        task = loop.create_task(worker.async_run())
        await asyncio.sleep(1)

        task.cancel()

        await worker.close()

        events = [item.payload for item in items]
        exception_event = events[0]
        assert exception_event["exception"]["values"][0]["type"] == "ZeroDivisionError"
        assert exception_event["transaction"] == "division"
    else:
        events = capture_events()

        loop = asyncio.get_event_loop()
        task = loop.create_task(worker.async_run())
        await asyncio.sleep(1)

        task.cancel()

        await worker.close()

        exception_event = events[1]
        assert exception_event["exception"]["values"][0]["type"] == "ZeroDivisionError"
        assert exception_event["transaction"] == "division"

    assert exception_event["extra"]["arq-job"]["task"] == "division"



from unittest import mock

import pytest

import sentry_sdk
from sentry_sdk.hub import Hub
from sentry_sdk.integrations import iter_default_integrations
from sentry_sdk.scrubber import DEFAULT_DENYLIST, EventScrubber

"""
Those tests are meant to check the compatibility of the new scopes in SDK 2.0 with the old Hub/Scope system in SDK 1.x.

Those tests have been run with the latest SDK 1.x version and the data used in the `assert` statements represents
the behvaior of the SDK 1.x.

This makes sure that we are backwards compatible. (on a best effort basis, there will probably be some edge cases that are not covered here)
"""


@pytest.fixture
def integrations():
    return [
        integration.identifier
        for integration in iter_default_integrations(
            with_auto_enabling_integrations=False
        )
    ]


@pytest.fixture
def expected_error(integrations):
    def create_expected_error_event(trx, span):
        return {
            "level": "warning-X",
            "exception": {
                "values": [
                    {
                        "mechanism": {"type": "generic", "handled": True},
                        "module": None,
                        "type": "ValueError",
                        "value": "This is a test exception",
                        "stacktrace": {
                            "frames": [
                                {
                                    "filename": "tests/new_scopes_compat/test_new_scopes_compat_event.py",
                                    "abs_path": mock.ANY,
                                    "function": "_faulty_function",
                                    "module": "tests.new_scopes_compat.test_new_scopes_compat_event",
                                    "lineno": mock.ANY,
                                    "pre_context": [
                                        "    return create_expected_transaction_event",
                                        "",
                                        "",
                                        "def _faulty_function():",
                                        "    try:",
                                    ],
                                    "context_line": '        raise ValueError("This is a test exception")',
                                    "post_context": [
                                        "    except ValueError as ex:",
                                        "        sentry_sdk.capture_exception(ex)",
                                        "",
                                        "",
                                        "def _test_before_send(event, hint):",
                                    ],
                                    "vars": {
                                        "ex": mock.ANY,
                                    },
                                    "in_app": True,
                                }
                            ]
                        },
                    }
                ]
            },
            "event_id": mock.ANY,
            "timestamp": mock.ANY,
            "contexts": {
                "character": {
                    "name": "Mighty Fighter changed by before_send",
                    "age": 19,
                    "attack_type": "melee",
                },
                "trace": {
                    "trace_id": trx.trace_id,
                    "span_id": span.span_id,
                    "parent_span_id": span.parent_span_id,
                    "op": "test_span",
                    "origin": "manual",
                    "description": None,
                    "data": {
                        "thread.id": mock.ANY,
                        "thread.name": "MainThread",
                    },
                },
                "runtime": {
                    "name": "CPython",
                    "version": mock.ANY,
                    "build": mock.ANY,
                },
            },
            "user": {
                "id": "123",
                "email": "jane.doe@example.com",
            },
            "transaction": "test_transaction",
            "transaction_info": {"source": "custom"},
            "tags": {"tag1": "tag1_value", "tag2": "tag2_value"},
            "extra": {
                "extra1": "extra1_value",
                "extra2": "extra2_value",
                "should_be_removed_by_event_scrubber": "[Filtered]",
                "sys.argv": "[Filtered]",
            },
            "breadcrumbs": {
                "values": [
                    {
                        "category": "error-level",
                        "message": "Authenticated user %s",
                        "level": "error",
                        "data": {"breadcrumb2": "somedata"},
                        "timestamp": mock.ANY,
                        "type": "default",
                    }
                ]
            },
            "modules": mock.ANY,
            "release": "0.1.2rc3",
            "environment": "checking-compatibility-with-sdk1",
            "server_name": mock.ANY,
            "sdk": {
                "name": "sentry.python",
                "version": mock.ANY,
                "packages": [{"name": "pypi:sentry-sdk", "version": mock.ANY}],
                "integrations": integrations,
            },
            "platform": "python",
            "_meta": {
                "extra": {
                    "should_be_removed_by_event_scrubber": {
                        "": {"rem": [["!config", "s"]]}
                    },
                    "sys.argv": {"": {"rem": [["!config", "s"]]}},
                },
            },
        }

    return create_expected_error_event


@pytest.fixture
def expected_transaction(integrations):
    def create_expected_transaction_event(trx, span):
        return {
            "type": "transaction",
            "transaction": "test_transaction changed by before_send_transaction",
            "transaction_info": {"source": "custom"},
            "contexts": {
                "trace": {
                    "trace_id": trx.trace_id,
                    "span_id": trx.span_id,
                    "parent_span_id": None,
                    "op": "test_transaction_op",
                    "origin": "manual",
                    "description": None,
                    "data": {
                        "thread.id": mock.ANY,
                        "thread.name": "MainThread",
                    },
                },
                "character": {
                    "name": "Mighty Fighter changed by before_send_transaction",
                    "age": 19,
                    "attack_type": "melee",
                },
                "runtime": {
                    "name": "CPython",
                    "version": mock.ANY,
                    "build": mock.ANY,
                },
            },
            "tags": {"tag1": "tag1_value", "tag2": "tag2_value"},
            "timestamp": mock.ANY,
            "start_timestamp": mock.ANY,
            "spans": [
                {
                    "data": {
                        "thread.id": mock.ANY,
                        "thread.name": "MainThread",
                    },
                    "trace_id": trx.trace_id,
                    "span_id": span.span_id,
                    "parent_span_id": span.parent_span_id,
                    "same_process_as_parent": True,
                    "op": "test_span",
                    "origin": "manual",
                    "description": None,
                    "start_timestamp": mock.ANY,
                    "timestamp": mock.ANY,
                }
            ],
            "measurements": {"memory_used": {"value": 456, "unit": "byte"}},
            "event_id": mock.ANY,
            "level": "warning-X",
            "user": {
                "id": "123",
                "email": "jane.doe@example.com",
            },
            "extra": {
                "extra1": "extra1_value",
                "extra2": "extra2_value",
                "should_be_removed_by_event_scrubber": "[Filtered]",
                "sys.argv": "[Filtered]",
            },
            "release": "0.1.2rc3",
            "environment": "checking-compatibility-with-sdk1",
            "server_name": mock.ANY,
            "sdk": {
                "name": "sentry.python",
                "version": mock.ANY,
                "packages": [{"name": "pypi:sentry-sdk", "version": mock.ANY}],
                "integrations": integrations,
            },
            "platform": "python",
            "_meta": {
                "extra": {
                    "should_be_removed_by_event_scrubber": {
                        "": {"rem": [["!config", "s"]]}
                    },
                    "sys.argv": {"": {"rem": [["!config", "s"]]}},
                },
            },
        }

    return create_expected_transaction_event


def _faulty_function():
    try:
        raise ValueError("This is a test exception")
    except ValueError as ex:
        sentry_sdk.capture_exception(ex)


def _test_before_send(event, hint):
    event["contexts"]["character"]["name"] += " changed by before_send"
    return event


def _test_before_send_transaction(event, hint):
    event["transaction"] += " changed by before_send_transaction"
    event["contexts"]["character"]["name"] += " changed by before_send_transaction"
    return event


def _test_before_breadcrumb(breadcrumb, hint):
    if breadcrumb["category"] == "info-level":
        return None
    return breadcrumb


def _generate_event_data(scope=None):
    """
    Generates some data to be used in the events sent by the tests.
    """
    sentry_sdk.set_level("warning-X")

    sentry_sdk.add_breadcrumb(
        category="info-level",
        message="Authenticated user %s",
        level="info",
        data={"breadcrumb1": "somedata"},
    )
    sentry_sdk.add_breadcrumb(
        category="error-level",
        message="Authenticated user %s",
        level="error",
        data={"breadcrumb2": "somedata"},
    )

    sentry_sdk.set_context(
        "character",
        {
            "name": "Mighty Fighter",
            "age": 19,
            "attack_type": "melee",
        },
    )

    sentry_sdk.set_extra("extra1", "extra1_value")
    sentry_sdk.set_extra("extra2", "extra2_value")
    sentry_sdk.set_extra("should_be_removed_by_event_scrubber", "XXX")

    sentry_sdk.set_tag("tag1", "tag1_value")
    sentry_sdk.set_tag("tag2", "tag2_value")

    sentry_sdk.set_user(
        {"id": "123", "email": "jane.doe@example.com", "ip_address": "211.161.1.124"}
    )

    sentry_sdk.set_measurement("memory_used", 456, "byte")

    if scope is not None:
        scope.add_attachment(bytes=b"Hello World", filename="hello.txt")


def _init_sentry_sdk(sentry_init):
    sentry_init(
        environment="checking-compatibility-with-sdk1",
        release="0.1.2rc3",
        before_send=_test_before_send,
        before_send_transaction=_test_before_send_transaction,
        before_breadcrumb=_test_before_breadcrumb,
        event_scrubber=EventScrubber(
            denylist=DEFAULT_DENYLIST
            + ["should_be_removed_by_event_scrubber", "sys.argv"]
        ),
        send_default_pii=False,
        traces_sample_rate=1.0,
        auto_enabling_integrations=False,
    )


#
# The actual Tests start here!
#


def test_event(sentry_init, capture_envelopes, expected_error, expected_transaction):
    _init_sentry_sdk(sentry_init)

    envelopes = capture_envelopes()

    with sentry_sdk.start_transaction(
        name="test_transaction", op="test_transaction_op"
    ) as trx:
        with sentry_sdk.start_span(op="test_span") as span:
            with sentry_sdk.configure_scope() as scope:  # configure scope
                _generate_event_data(scope)
                _faulty_function()

    (error_envelope, transaction_envelope) = envelopes

    error = error_envelope.get_event()
    transaction = transaction_envelope.get_transaction_event()
    attachment = error_envelope.items[-1]

    assert error == expected_error(trx, span)
    assert transaction == expected_transaction(trx, span)
    assert attachment.headers == {
        "filename": "hello.txt",
        "type": "attachment",
        "content_type": "text/plain",
    }
    assert attachment.payload.bytes == b"Hello World"


def test_event2(sentry_init, capture_envelopes, expected_error, expected_transaction):
    _init_sentry_sdk(sentry_init)

    envelopes = capture_envelopes()

    with Hub(Hub.current):
        sentry_sdk.set_tag("A", 1)  # will not be added

    with Hub.current:  # with hub
        with sentry_sdk.push_scope() as scope:
            scope.set_tag("B", 1)  # will not be added

        with sentry_sdk.start_transaction(
            name="test_transaction", op="test_transaction_op"
        ) as trx:
            with sentry_sdk.start_span(op="test_span") as span:
                with sentry_sdk.configure_scope() as scope:  # configure scope
                    _generate_event_data(scope)
                    _faulty_function()

    (error_envelope, transaction_envelope) = envelopes

    error = error_envelope.get_event()
    transaction = transaction_envelope.get_transaction_event()
    attachment = error_envelope.items[-1]

    assert error == expected_error(trx, span)
    assert transaction == expected_transaction(trx, span)
    assert attachment.headers == {
        "filename": "hello.txt",
        "type": "attachment",
        "content_type": "text/plain",
    }
    assert attachment.payload.bytes == b"Hello World"


def test_event3(sentry_init, capture_envelopes, expected_error, expected_transaction):
    _init_sentry_sdk(sentry_init)

    envelopes = capture_envelopes()

    with Hub(Hub.current):
        sentry_sdk.set_tag("A", 1)  # will not be added

    with Hub.current:  # with hub
        with sentry_sdk.push_scope() as scope:
            scope.set_tag("B", 1)  # will not be added

        with sentry_sdk.push_scope() as scope:  # push scope
            with sentry_sdk.start_transaction(
                name="test_transaction", op="test_transaction_op"
            ) as trx:
                with sentry_sdk.start_span(op="test_span") as span:
                    _generate_event_data(scope)
                    _faulty_function()

    (error_envelope, transaction_envelope) = envelopes

    error = error_envelope.get_event()
    transaction = transaction_envelope.get_transaction_event()
    attachment = error_envelope.items[-1]

    assert error == expected_error(trx, span)
    assert transaction == expected_transaction(trx, span)
    assert attachment.headers == {
        "filename": "hello.txt",
        "type": "attachment",
        "content_type": "text/plain",
    }
    assert attachment.payload.bytes == b"Hello World"


def test_event4(sentry_init, capture_envelopes, expected_error, expected_transaction):
    _init_sentry_sdk(sentry_init)

    envelopes = capture_envelopes()

    with Hub(Hub.current):
        sentry_sdk.set_tag("A", 1)  # will not be added

    with Hub(Hub.current):  # with hub clone
        with sentry_sdk.push_scope() as scope:
            scope.set_tag("B", 1)  # will not be added

        with sentry_sdk.start_transaction(
            name="test_transaction", op="test_transaction_op"
        ) as trx:
            with sentry_sdk.start_span(op="test_span") as span:
                with sentry_sdk.configure_scope() as scope:  # configure scope
                    _generate_event_data(scope)
                    _faulty_function()

    (error_envelope, transaction_envelope) = envelopes

    error = error_envelope.get_event()
    transaction = transaction_envelope.get_transaction_event()
    attachment = error_envelope.items[-1]

    assert error == expected_error(trx, span)
    assert transaction == expected_transaction(trx, span)
    assert attachment.headers == {
        "filename": "hello.txt",
        "type": "attachment",
        "content_type": "text/plain",
    }
    assert attachment.payload.bytes == b"Hello World"


def test_event5(sentry_init, capture_envelopes, expected_error, expected_transaction):
    _init_sentry_sdk(sentry_init)

    envelopes = capture_envelopes()

    with Hub(Hub.current):
        sentry_sdk.set_tag("A", 1)  # will not be added

    with Hub(Hub.current):  # with hub clone
        with sentry_sdk.push_scope() as scope:
            scope.set_tag("B", 1)  # will not be added

        with sentry_sdk.push_scope() as scope:  # push scope
            with sentry_sdk.start_transaction(
                name="test_transaction", op="test_transaction_op"
            ) as trx:
                with sentry_sdk.start_span(op="test_span") as span:
                    _generate_event_data(scope)
                    _faulty_function()

    (error_envelope, transaction_envelope) = envelopes

    error = error_envelope.get_event()
    transaction = transaction_envelope.get_transaction_event()
    attachment = error_envelope.items[-1]

    assert error == expected_error(trx, span)
    assert transaction == expected_transaction(trx, span)
    assert attachment.headers == {
        "filename": "hello.txt",
        "type": "attachment",
        "content_type": "text/plain",
    }
    assert attachment.payload.bytes == b"Hello World"



import uuid
from unittest import mock

import pytest

import sentry_sdk
from sentry_sdk.crons import capture_checkin


@sentry_sdk.monitor(monitor_slug="abc123")
def _hello_world(name):
    return "Hello, {}".format(name)


@sentry_sdk.monitor(monitor_slug="def456")
def _break_world(name):
    1 / 0
    return "Hello, {}".format(name)


def _hello_world_contextmanager(name):
    with sentry_sdk.monitor(monitor_slug="abc123"):
        return "Hello, {}".format(name)


def _break_world_contextmanager(name):
    with sentry_sdk.monitor(monitor_slug="def456"):
        1 / 0
        return "Hello, {}".format(name)


@sentry_sdk.monitor(monitor_slug="abc123")
async def _hello_world_async(name):
    return "Hello, {}".format(name)


@sentry_sdk.monitor(monitor_slug="def456")
async def _break_world_async(name):
    1 / 0
    return "Hello, {}".format(name)


async def my_coroutine():
    return


async def _hello_world_contextmanager_async(name):
    with sentry_sdk.monitor(monitor_slug="abc123"):
        await my_coroutine()
        return "Hello, {}".format(name)


async def _break_world_contextmanager_async(name):
    with sentry_sdk.monitor(monitor_slug="def456"):
        await my_coroutine()
        1 / 0
        return "Hello, {}".format(name)


@sentry_sdk.monitor(monitor_slug="ghi789", monitor_config=None)
def _no_monitor_config():
    return


@sentry_sdk.monitor(
    monitor_slug="ghi789",
    monitor_config={
        "schedule": {"type": "crontab", "value": "0 0 * * *"},
        "failure_issue_threshold": 5,
    },
)
def _with_monitor_config():
    return


def test_decorator(sentry_init):
    sentry_init()

    with mock.patch(
        "sentry_sdk.crons.decorator.capture_checkin"
    ) as fake_capture_checkin:
        result = _hello_world("Grace")
        assert result == "Hello, Grace"

        # Check for initial checkin
        fake_capture_checkin.assert_has_calls(
            [
                mock.call(
                    monitor_slug="abc123", status="in_progress", monitor_config=None
                ),
            ]
        )

        # Check for final checkin
        assert fake_capture_checkin.call_args[1]["monitor_slug"] == "abc123"
        assert fake_capture_checkin.call_args[1]["status"] == "ok"
        assert fake_capture_checkin.call_args[1]["duration"]
        assert fake_capture_checkin.call_args[1]["check_in_id"]


def test_decorator_error(sentry_init):
    sentry_init()

    with mock.patch(
        "sentry_sdk.crons.decorator.capture_checkin"
    ) as fake_capture_checkin:
        with pytest.raises(ZeroDivisionError):
            result = _break_world("Grace")

        assert "result" not in locals()

        # Check for initial checkin
        fake_capture_checkin.assert_has_calls(
            [
                mock.call(
                    monitor_slug="def456", status="in_progress", monitor_config=None
                ),
            ]
        )

        # Check for final checkin
        assert fake_capture_checkin.call_args[1]["monitor_slug"] == "def456"
        assert fake_capture_checkin.call_args[1]["status"] == "error"
        assert fake_capture_checkin.call_args[1]["duration"]
        assert fake_capture_checkin.call_args[1]["check_in_id"]


def test_contextmanager(sentry_init):
    sentry_init()

    with mock.patch(
        "sentry_sdk.crons.decorator.capture_checkin"
    ) as fake_capture_checkin:
        result = _hello_world_contextmanager("Grace")
        assert result == "Hello, Grace"

        # Check for initial checkin
        fake_capture_checkin.assert_has_calls(
            [
                mock.call(
                    monitor_slug="abc123", status="in_progress", monitor_config=None
                ),
            ]
        )

        # Check for final checkin
        assert fake_capture_checkin.call_args[1]["monitor_slug"] == "abc123"
        assert fake_capture_checkin.call_args[1]["status"] == "ok"
        assert fake_capture_checkin.call_args[1]["duration"]
        assert fake_capture_checkin.call_args[1]["check_in_id"]


def test_contextmanager_error(sentry_init):
    sentry_init()

    with mock.patch(
        "sentry_sdk.crons.decorator.capture_checkin"
    ) as fake_capture_checkin:
        with pytest.raises(ZeroDivisionError):
            result = _break_world_contextmanager("Grace")

        assert "result" not in locals()

        # Check for initial checkin
        fake_capture_checkin.assert_has_calls(
            [
                mock.call(
                    monitor_slug="def456", status="in_progress", monitor_config=None
                ),
            ]
        )

        # Check for final checkin
        assert fake_capture_checkin.call_args[1]["monitor_slug"] == "def456"
        assert fake_capture_checkin.call_args[1]["status"] == "error"
        assert fake_capture_checkin.call_args[1]["duration"]
        assert fake_capture_checkin.call_args[1]["check_in_id"]


def test_capture_checkin_simple(sentry_init):
    sentry_init()

    check_in_id = capture_checkin(
        monitor_slug="abc123",
        check_in_id="112233",
        status=None,
        duration=None,
    )
    assert check_in_id == "112233"


def test_sample_rate_doesnt_affect_crons(sentry_init, capture_envelopes):
    sentry_init(sample_rate=0)
    envelopes = capture_envelopes()

    capture_checkin(check_in_id="112233")

    assert len(envelopes) == 1

    check_in = envelopes[0].items[0].payload.json
    assert check_in["check_in_id"] == "112233"


def test_capture_checkin_new_id(sentry_init):
    sentry_init()

    with mock.patch("uuid.uuid4") as mock_uuid:
        mock_uuid.return_value = uuid.UUID("a8098c1a-f86e-11da-bd1a-00112444be1e")
        check_in_id = capture_checkin(
            monitor_slug="abc123",
            check_in_id=None,
            status=None,
            duration=None,
        )

        assert check_in_id == "a8098c1af86e11dabd1a00112444be1e"


def test_end_to_end(sentry_init, capture_envelopes):
    sentry_init()
    envelopes = capture_envelopes()

    capture_checkin(
        monitor_slug="abc123",
        check_in_id="112233",
        duration=123,
        status="ok",
    )

    check_in = envelopes[0].items[0].payload.json

    # Check for final checkin
    assert check_in["check_in_id"] == "112233"
    assert check_in["monitor_slug"] == "abc123"
    assert check_in["status"] == "ok"
    assert check_in["duration"] == 123


def test_monitor_config(sentry_init, capture_envelopes):
    sentry_init()
    envelopes = capture_envelopes()

    monitor_config = {
        "schedule": {"type": "crontab", "value": "0 0 * * *"},
        "failure_issue_threshold": 5,
        "recovery_threshold": 5,
    }

    capture_checkin(monitor_slug="abc123", monitor_config=monitor_config)
    check_in = envelopes[0].items[0].payload.json

    # Check for final checkin
    assert check_in["monitor_slug"] == "abc123"
    assert check_in["monitor_config"] == monitor_config

    # Without passing a monitor_config the field is not in the checkin
    capture_checkin(monitor_slug="abc123")
    check_in = envelopes[1].items[0].payload.json

    assert check_in["monitor_slug"] == "abc123"
    assert "monitor_config" not in check_in


def test_monitor_config_with_owner(sentry_init, capture_envelopes):
    sentry_init()
    envelopes = capture_envelopes()

    monitor_config = {
        "schedule": {"type": "crontab", "value": "0 0 * * *"},
        "owner": "team:6",
    }

    capture_checkin(monitor_slug="abc123", monitor_config=monitor_config)
    check_in = envelopes[0].items[0].payload.json

    assert check_in["monitor_slug"] == "abc123"
    assert check_in["monitor_config"]["owner"] == "team:6"


def test_decorator_monitor_config(sentry_init, capture_envelopes):
    sentry_init()
    envelopes = capture_envelopes()

    _with_monitor_config()

    assert len(envelopes) == 2

    for check_in_envelope in envelopes:
        assert len(check_in_envelope.items) == 1
        check_in = check_in_envelope.items[0].payload.json

        assert check_in["monitor_slug"] == "ghi789"
        assert check_in["monitor_config"] == {
            "schedule": {"type": "crontab", "value": "0 0 * * *"},
            "failure_issue_threshold": 5,
        }


def test_decorator_no_monitor_config(sentry_init, capture_envelopes):
    sentry_init()
    envelopes = capture_envelopes()

    _no_monitor_config()

    assert len(envelopes) == 2

    for check_in_envelope in envelopes:
        assert len(check_in_envelope.items) == 1
        check_in = check_in_envelope.items[0].payload.json

        assert check_in["monitor_slug"] == "ghi789"
        assert "monitor_config" not in check_in


def test_capture_checkin_sdk_not_initialized():
    # Tests that the capture_checkin does not raise an error when Sentry SDK is not initialized.
    # sentry_init() is intentionally omitted.
    check_in_id = capture_checkin(
        monitor_slug="abc123",
        check_in_id="112233",
        status=None,
        duration=None,
    )
    assert check_in_id == "112233"


def test_scope_data_in_checkin(sentry_init, capture_envelopes):
    sentry_init()
    envelopes = capture_envelopes()

    valid_keys = [
        # Mandatory event keys
        "type",
        "event_id",
        "timestamp",
        "platform",
        # Optional event keys
        "release",
        "environment",
        "server_name",
        "sdk",
        # Mandatory check-in specific keys
        "check_in_id",
        "monitor_slug",
        "status",
        # Optional check-in specific keys
        "duration",
        "monitor_config",
        "contexts",  # an event processor adds this
    ]

    # Add some data to the scope
    sentry_sdk.add_breadcrumb(message="test breadcrumb")
    sentry_sdk.set_context("test_context", {"test_key": "test_value"})
    sentry_sdk.set_extra("test_extra", "test_value")
    sentry_sdk.set_level("warning")
    sentry_sdk.set_tag("test_tag", "test_value")

    capture_checkin(
        monitor_slug="abc123",
        check_in_id="112233",
        status="ok",
        duration=123,
    )

    (envelope,) = envelopes
    check_in_event = envelope.items[0].payload.json

    invalid_keys = []
    for key in check_in_event.keys():
        if key not in valid_keys:
            invalid_keys.append(key)

    assert len(invalid_keys) == 0, "Unexpected keys found in checkin: {}".format(
        invalid_keys
    )


@pytest.mark.asyncio
async def test_decorator_async(sentry_init):
    sentry_init()

    with mock.patch(
        "sentry_sdk.crons.decorator.capture_checkin"
    ) as fake_capture_checkin:
        result = await _hello_world_async("Grace")
        assert result == "Hello, Grace"

        # Check for initial checkin
        fake_capture_checkin.assert_has_calls(
            [
                mock.call(
                    monitor_slug="abc123", status="in_progress", monitor_config=None
                ),
            ]
        )

        # Check for final checkin
        assert fake_capture_checkin.call_args[1]["monitor_slug"] == "abc123"
        assert fake_capture_checkin.call_args[1]["status"] == "ok"
        assert fake_capture_checkin.call_args[1]["duration"]
        assert fake_capture_checkin.call_args[1]["check_in_id"]


@pytest.mark.asyncio
async def test_decorator_error_async(sentry_init):
    sentry_init()

    with mock.patch(
        "sentry_sdk.crons.decorator.capture_checkin"
    ) as fake_capture_checkin:
        with pytest.raises(ZeroDivisionError):
            result = await _break_world_async("Grace")

        assert "result" not in locals()

        # Check for initial checkin
        fake_capture_checkin.assert_has_calls(
            [
                mock.call(
                    monitor_slug="def456", status="in_progress", monitor_config=None
                ),
            ]
        )

        # Check for final checkin
        assert fake_capture_checkin.call_args[1]["monitor_slug"] == "def456"
        assert fake_capture_checkin.call_args[1]["status"] == "error"
        assert fake_capture_checkin.call_args[1]["duration"]
        assert fake_capture_checkin.call_args[1]["check_in_id"]


@pytest.mark.asyncio
async def test_contextmanager_async(sentry_init):
    sentry_init()

    with mock.patch(
        "sentry_sdk.crons.decorator.capture_checkin"
    ) as fake_capture_checkin:
        result = await _hello_world_contextmanager_async("Grace")
        assert result == "Hello, Grace"

        # Check for initial checkin
        fake_capture_checkin.assert_has_calls(
            [
                mock.call(
                    monitor_slug="abc123", status="in_progress", monitor_config=None
                ),
            ]
        )

        # Check for final checkin
        assert fake_capture_checkin.call_args[1]["monitor_slug"] == "abc123"
        assert fake_capture_checkin.call_args[1]["status"] == "ok"
        assert fake_capture_checkin.call_args[1]["duration"]
        assert fake_capture_checkin.call_args[1]["check_in_id"]


@pytest.mark.asyncio
async def test_contextmanager_error_async(sentry_init):
    sentry_init()

    with mock.patch(
        "sentry_sdk.crons.decorator.capture_checkin"
    ) as fake_capture_checkin:
        with pytest.raises(ZeroDivisionError):
            result = await _break_world_contextmanager_async("Grace")

        assert "result" not in locals()

        # Check for initial checkin
        fake_capture_checkin.assert_has_calls(
            [
                mock.call(
                    monitor_slug="def456", status="in_progress", monitor_config=None
                ),
            ]
        )

        # Check for final checkin
        assert fake_capture_checkin.call_args[1]["monitor_slug"] == "def456"
        assert fake_capture_checkin.call_args[1]["status"] == "error"
        assert fake_capture_checkin.call_args[1]["duration"]
        assert fake_capture_checkin.call_args[1]["check_in_id"]



import sys

import pytest

from sentry_sdk.types import Event, Hint


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason="Type hinting with `|` is available in Python 3.10+",
)
def test_event_or_none_runtime():
    """
    Ensures that the `Event` type's runtime value supports the `|` operation with `None`.
    This test is needed to ensure that using an `Event | None` type hint (e.g. for
    `before_send`'s return value) does not raise a TypeError at runtime.
    """
    Event | None


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason="Type hinting with `|` is available in Python 3.10+",
)
def test_hint_or_none_runtime():
    """
    Analogue to `test_event_or_none_runtime`, but for the `Hint` type.
    """
    Hint | None



import json
import os
import shutil
import time
import uuid

import pytest
import ray

import sentry_sdk
from sentry_sdk.envelope import Envelope
from sentry_sdk.integrations.ray import RayIntegration
from sentry_sdk.integrations.stdlib import StdlibIntegration
from sentry_sdk.traces import SegmentNameSource
from tests.conftest import TestTransport


@pytest.fixture(autouse=True)
def shutdown_ray(tmpdir):
    yield
    ray.shutdown()


class RayTestTransport(TestTransport):
    def __init__(self):
        self.envelopes = []
        super().__init__()

    def capture_envelope(self, envelope: Envelope) -> None:
        self.envelopes.append(envelope)


class RayLoggingTransport(TestTransport):
    def capture_envelope(self, envelope: Envelope) -> None:
        print(envelope.serialize().decode("utf-8", "replace"))


def setup_sentry_with_logging_transport():
    setup_sentry(transport=RayLoggingTransport())


def setup_sentry_with_logging_transport_and_span_streaming():
    setup_sentry(span_streaming=True, transport=RayLoggingTransport())


def setup_sentry(span_streaming=False, transport=None):
    if span_streaming:
        sentry_sdk._span_batcher.SpanBatcher.MAX_BEFORE_FLUSH = 1

    sentry_sdk.init(
        integrations=[RayIntegration()],
        transport=RayTestTransport() if transport is None else transport,
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )


def read_error_from_log(job_id, ray_temp_dir):
    # Find the actual session directory that Ray created
    session_dirs = [d for d in os.listdir(ray_temp_dir) if d.startswith("session_")]
    if not session_dirs:
        raise FileNotFoundError(f"No session directory found in {ray_temp_dir}")

    session_dir = os.path.join(ray_temp_dir, session_dirs[0])
    log_dir = os.path.join(session_dir, "logs")

    if not os.path.exists(log_dir):
        raise FileNotFoundError(f"No logs directory found at {log_dir}")

    log_file = [
        f
        for f in os.listdir(log_dir)
        if "worker" in f and job_id in f and f.endswith(".out")
    ][0]

    next_line_is_event_payload = False
    with open(os.path.join(log_dir, log_file), "r") as file:
        for line in file:
            try:
                payload = json.loads(line)
            except ValueError:
                continue

            if next_line_is_event_payload:
                return payload

            if isinstance(payload, dict) and payload.get("type") == "event":
                next_line_is_event_payload = True

    return None


def _parse_spans_from_log(job_id, ray_temp_dir):
    # Find the actual session directory that Ray created
    session_dirs = [d for d in os.listdir(ray_temp_dir) if d.startswith("session_")]
    if not session_dirs:
        raise FileNotFoundError(f"No session directory found in {ray_temp_dir}")

    session_dir = os.path.join(ray_temp_dir, session_dirs[0])
    log_dir = os.path.join(session_dir, "logs")

    if not os.path.exists(log_dir):
        raise FileNotFoundError(f"No logs directory found at {log_dir}")

    log_files = [
        f
        for f in os.listdir(log_dir)
        if "worker" in f and job_id in f and f.endswith(".out")
    ]
    if not log_files:
        return []

    spans = []
    next_line_is_span_payload = False
    with open(os.path.join(log_dir, log_files[0]), "r") as file:
        for line in file:
            try:
                payload = json.loads(line)
            except ValueError:
                continue
            if next_line_is_span_payload:
                spans.extend(payload["items"])
                next_line_is_span_payload = False
                continue
            if isinstance(payload, dict) and payload.get("type") == "span":
                next_line_is_span_payload = True

    return spans


def read_spans_from_log(job_id, ray_temp_dir, min_spans=1, timeout=10):
    deadline = time.monotonic() + timeout
    spans = []
    while True:
        try:
            spans = _parse_spans_from_log(job_id, ray_temp_dir)
        except FileNotFoundError:
            spans = []
        if len(spans) >= min_spans or time.monotonic() >= deadline:
            return spans
        time.sleep(0.1)


def example_task(span_streaming: bool):
    if span_streaming:
        with sentry_sdk.traces.start_span(
            name="example task step",
            attributes={
                "sentry.op": "task",
            },
        ):
            ...
    else:
        with sentry_sdk.start_span(op="task", name="example task step"):
            ...

        return sentry_sdk.get_client().transport.envelopes


# RayIntegration must leave variadic keyword arguments at the end
def example_task_with_kwargs(span_streaming: bool, **kwargs):
    if span_streaming:
        with sentry_sdk.traces.start_span(
            name="example task step", attributes={"sentry.op": "task"}
        ):
            ...
    else:
        with sentry_sdk.start_span(op="task", name="example task step"):
            ...

        return sentry_sdk.get_client().transport.envelopes


@pytest.mark.parametrize(
    "task_options", [{}, {"num_cpus": 0, "memory": 1024 * 1024 * 10}]
)
@pytest.mark.parametrize(
    "task",
    [example_task, example_task_with_kwargs],
)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_tracing_in_ray_tasks(task_options, task, span_streaming):
    sentry_sdk.init(
        integrations=[RayIntegration()],
        disabled_integrations=[StdlibIntegration],
        transport=RayTestTransport(),
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    # Setup ray task, calling decorator directly instead of @,
    # to accommodate for test parametrization
    if task_options:
        example_task = ray.remote(**task_options)(task)
    else:
        example_task = ray.remote(task)

    # Function name shouldn't be overwritten by Sentry wrapper
    assert (
        example_task._function_name
        == f"tests.integrations.ray.test_ray.{task.__name__}"
    )

    if span_streaming:
        ray_temp_dir = os.path.join("/tmp", f"ray_test_{uuid.uuid4().hex[:8]}")
        os.makedirs(ray_temp_dir, exist_ok=True)

        try:
            ray.init(
                runtime_env={
                    "worker_process_setup_hook": setup_sentry_with_logging_transport_and_span_streaming,
                    "working_dir": "./",
                },
                _temp_dir=ray_temp_dir,
            )

            with sentry_sdk.traces.start_span(
                name="ray test parent", attributes={"sentry.op": "task"}
            ):
                future = example_task.remote(span_streaming)
                ray.get(future)

            job_id = future.job_id().hex()
            worker_spans = read_spans_from_log(job_id, ray_temp_dir, min_spans=2)
        finally:
            if os.path.exists(ray_temp_dir):
                shutil.rmtree(ray_temp_dir, ignore_errors=True)

        sentry_sdk.flush()
        client_envelope = sentry_sdk.get_client().transport.envelopes[0]
        client_spans = [
            span
            for item in client_envelope.items
            for span in item.payload.json["items"]
        ]
        assert client_spans[1]["name"] == "ray test parent"

        assert (
            worker_spans[1]["name"]
            == f"tests.integrations.ray.test_ray.{task.__name__}"
        )
        assert (
            worker_spans[1]["attributes"]["sentry.segment.name.source"]["value"]
            == SegmentNameSource.TASK
        )

        span = client_spans[0]
        assert span["attributes"]["sentry.op"]["value"] == "queue.submit.ray"
        assert span["attributes"]["sentry.origin"]["value"] == "auto.queue.ray"
        assert span["name"] == f"tests.integrations.ray.test_ray.{task.__name__}"
        assert span["parent_span_id"] == client_spans[1]["span_id"]
        assert span["trace_id"] == client_spans[1]["trace_id"]

        span = worker_spans[0]
        assert span["attributes"]["sentry.op"]["value"] == "task"
        assert span["attributes"]["sentry.origin"]["value"] == "manual"
        assert span["name"] == "example task step"
        assert span["parent_span_id"] == worker_spans[1]["span_id"]
        assert span["trace_id"] == worker_spans[1]["trace_id"]

        assert client_spans[1]["trace_id"] == worker_spans[1]["trace_id"]
    else:
        ray.init(
            runtime_env={
                "worker_process_setup_hook": setup_sentry,
                "working_dir": "./",
            }
        )

        with sentry_sdk.start_transaction(op="task", name="ray test transaction"):
            worker_envelopes = ray.get(example_task.remote(span_streaming))

        client_envelope = sentry_sdk.get_client().transport.envelopes[0]
        client_transaction = client_envelope.get_transaction_event()
        assert client_transaction["transaction"] == "ray test transaction"
        assert client_transaction["transaction_info"] == {"source": "custom"}

        worker_envelope = worker_envelopes[0]
        worker_transaction = worker_envelope.get_transaction_event()
        assert (
            worker_transaction["transaction"]
            == f"tests.integrations.ray.test_ray.{task.__name__}"
        )
        assert worker_transaction["transaction_info"] == {"source": "task"}

        (span,) = client_transaction["spans"]
        assert span["op"] == "queue.submit.ray"
        assert span["origin"] == "auto.queue.ray"
        assert span["description"] == f"tests.integrations.ray.test_ray.{task.__name__}"
        assert (
            span["parent_span_id"] == client_transaction["contexts"]["trace"]["span_id"]
        )
        assert span["trace_id"] == client_transaction["contexts"]["trace"]["trace_id"]

        (span,) = worker_transaction["spans"]
        assert span["op"] == "task"
        assert span["origin"] == "manual"
        assert span["description"] == "example task step"
        assert (
            span["parent_span_id"] == worker_transaction["contexts"]["trace"]["span_id"]
        )
        assert span["trace_id"] == worker_transaction["contexts"]["trace"]["trace_id"]

        assert (
            client_transaction["contexts"]["trace"]["trace_id"]
            == worker_transaction["contexts"]["trace"]["trace_id"]
        )


@pytest.mark.parametrize("span_streaming", [True, False])
def test_errors_in_ray_tasks(span_streaming):
    sentry_sdk.init(
        integrations=[RayIntegration()],
        disabled_integrations=[StdlibIntegration],
        transport=RayTestTransport(),
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    ray_temp_dir = os.path.join("/tmp", f"ray_test_{uuid.uuid4().hex[:8]}")
    os.makedirs(ray_temp_dir, exist_ok=True)

    try:
        ray.init(
            runtime_env={
                "worker_process_setup_hook": setup_sentry_with_logging_transport_and_span_streaming
                if span_streaming
                else setup_sentry_with_logging_transport,
                "working_dir": "./",
            },
            _temp_dir=ray_temp_dir,
        )

        # Setup ray task
        @ray.remote
        def example_task():
            1 / 0

        if span_streaming:
            with sentry_sdk.traces.start_span(
                name="ray test parent", attributes={"sentry.op": "task"}
            ):
                with pytest.raises(ZeroDivisionError):
                    future = example_task.remote()
                    ray.get(future)
        else:
            with sentry_sdk.start_transaction(op="task", name="ray test transaction"):
                with pytest.raises(ZeroDivisionError):
                    future = example_task.remote()
                    ray.get(future)

        job_id = future.job_id().hex()
        error = read_error_from_log(job_id, ray_temp_dir)

        assert error["level"] == "error"
        assert (
            error["transaction"]
            == "tests.integrations.ray.test_ray.test_errors_in_ray_tasks.<locals>.example_task"
        )
        assert error["exception"]["values"][0]["mechanism"]["type"] == "ray"
        assert not error["exception"]["values"][0]["mechanism"]["handled"]

    finally:
        if os.path.exists(ray_temp_dir):
            shutil.rmtree(ray_temp_dir, ignore_errors=True)


# Arbitrary keyword argument to test all decorator paths
@pytest.mark.parametrize("remote_kwargs", [{}, {"namespace": "actors"}])
@pytest.mark.parametrize("span_streaming", [True, False])
def test_tracing_in_ray_actors(remote_kwargs, span_streaming):
    sentry_sdk.init(
        integrations=[RayIntegration()],
        disabled_integrations=[StdlibIntegration],
        transport=RayTestTransport(),
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    # Setup ray actor
    if remote_kwargs:

        @ray.remote(**remote_kwargs)
        class Counter:
            def __init__(self):
                self.n = 0

            def increment(self):
                if span_streaming:
                    with sentry_sdk.traces.start_span(
                        name="example actor execution", attributes={"sentry.op": "task"}
                    ):
                        self.n += 1
                else:
                    with sentry_sdk.start_span(
                        op="task", name="example actor execution"
                    ):
                        self.n += 1

                    return sentry_sdk.get_client().transport.envelopes
    else:

        @ray.remote
        class Counter:
            def __init__(self):
                self.n = 0

            def increment(self):
                if span_streaming:
                    with sentry_sdk.traces.start_span(
                        name="example actor execution", attributes={"sentry.op": "task"}
                    ):
                        self.n += 1
                else:
                    with sentry_sdk.start_span(
                        op="task", name="example actor execution"
                    ):
                        self.n += 1

                    return sentry_sdk.get_client().transport.envelopes

    if span_streaming:
        ray_temp_dir = os.path.join("/tmp", f"ray_test_{uuid.uuid4().hex[:8]}")
        os.makedirs(ray_temp_dir, exist_ok=True)

        try:
            ray.init(
                runtime_env={
                    "worker_process_setup_hook": setup_sentry_with_logging_transport_and_span_streaming,
                    "working_dir": "./",
                },
                _temp_dir=ray_temp_dir,
            )

            with sentry_sdk.traces.start_span(
                name="ray test parent", attributes={"sentry.op": "task"}
            ):
                counter = Counter.remote()
                future = counter.increment.remote()
                ray.get(future)

            job_id = future.job_id().hex()
            worker_spans = read_spans_from_log(job_id, ray_temp_dir)
        finally:
            if os.path.exists(ray_temp_dir):
                shutil.rmtree(ray_temp_dir, ignore_errors=True)

        sentry_sdk.flush()
        client_envelope = sentry_sdk.get_client().transport.envelopes[0]
        client_spans = [
            span
            for item in client_envelope.items
            for span in item.payload.json["items"]
        ]

        # Spans for submitting the actor task are not created (actors are not supported yet)
        # Only the manual "example actor execution" span is recorded.
        assert len(client_spans) == 1

        # Transaction are not yet created when executing ray actors (actors are not supported yet)
        # Only the manual "example actor execution" span is recorded.
        assert len(worker_spans) == 1
    else:
        ray.init(
            runtime_env={
                "worker_process_setup_hook": setup_sentry,
                "working_dir": "./",
            }
        )

        with sentry_sdk.start_transaction(op="task", name="ray test transaction"):
            counter = Counter.remote()
            worker_envelopes = ray.get(counter.increment.remote())

        client_envelope = sentry_sdk.get_client().transport.envelopes[0]
        client_transaction = client_envelope.get_transaction_event()

        # Spans for submitting the actor task are not created (actors are not supported yet)
        assert client_transaction["spans"] == []

        # Transaction are not yet created when executing ray actors (actors are not supported yet)
        assert worker_envelopes == []


@pytest.mark.parametrize("span_streaming", [True, False])
def test_errors_in_ray_actors(span_streaming):
    sentry_sdk.init(
        integrations=[RayIntegration()],
        disabled_integrations=[StdlibIntegration],
        transport=RayLoggingTransport(),
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    ray_temp_dir = os.path.join("/tmp", f"ray_test_{uuid.uuid4().hex[:8]}")
    os.makedirs(ray_temp_dir, exist_ok=True)

    try:
        ray.init(
            runtime_env={
                "worker_process_setup_hook": setup_sentry_with_logging_transport_and_span_streaming
                if span_streaming
                else setup_sentry_with_logging_transport,
                "working_dir": "./",
            },
            _temp_dir=ray_temp_dir,
        )

        # Setup ray actor
        @ray.remote
        class Counter:
            def __init__(self):
                self.n = 0

            def increment(self):
                if span_streaming:
                    with sentry_sdk.traces.start_span(
                        name="example actor execution", attributes={"sentry.op": "task"}
                    ):
                        1 / 0
                else:
                    with sentry_sdk.start_span(
                        op="task", name="example actor execution"
                    ):
                        1 / 0

                return sentry_sdk.get_client().transport.envelopes

        if span_streaming:
            with sentry_sdk.traces.start_span(
                name="ray test parent", attributes={"sentry.op": "task"}
            ):
                with pytest.raises(ZeroDivisionError):
                    counter = Counter.remote()
                    future = counter.increment.remote()
                    ray.get(future)
        else:
            with sentry_sdk.start_transaction(op="task", name="ray test transaction"):
                with pytest.raises(ZeroDivisionError):
                    counter = Counter.remote()
                    future = counter.increment.remote()
                    ray.get(future)

        job_id = future.job_id().hex()
        error = read_error_from_log(job_id, ray_temp_dir)

        # We do not capture errors in ray actors yet
        assert error is None

    finally:
        if os.path.exists(ray_temp_dir):
            shutil.rmtree(ray_temp_dir, ignore_errors=True)
